@@ -3,37 +3,85 @@
  derived from the source code of Encryption for the Masses 2.02a, which is
  Copyright (c) 1998-2000 Paul Le Roux and which is governed by the 'License
  Agreement for Encryption for the Masses'. Modifications and additions to
- the original source code (contained in this file) and all other portions of
- this file are Copyright (c) 2003-2008 TrueCrypt Foundation and are governed
- by the TrueCrypt License 2.4 the full text of which is contained in the
- file License.txt included in TrueCrypt binary and source code distribution
- packages. */
+ the original source code (contained in this file) and all other portions
+ of this file are Copyright (c) 2003-2010 TrueCrypt Developers Association
+ and are governed by the TrueCrypt License 3.0 the full text of which is
+ contained in the file License.txt included in TrueCrypt binary and source
+ code distribution packages. */
 
 #include "Tcdefs.h"
 
 #ifndef TC_WINDOWS_BOOT
 #include <fcntl.h>
-#include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <time.h>
+#include "EncryptionThreadPool.h"
 #endif
 
-#ifdef _WIN32
+#include <stddef.h>
+#include <string.h>
 #include <io.h>
+
+#ifndef DEVICE_DRIVER
 #include "Random.h"
 #endif
 
-#include "Crypto.h"
-#include "Common/Endian.h"
-#include "Volumes.h"
-
-#include "Pkcs5.h"
 #include "Crc.h"
+#include "Crypto.h"
+#include "Endian.h"
+#include "Volumes.h"
+#include "Pkcs5.h"
 
 
+/* Volume header v5 structure (used since TrueCrypt 7.0): */
+//
+// Offset	Length	Description
+// ------------------------------------------
+// Unencrypted:
+// 0		64		Salt
+// Encrypted:
+// 64		4		ASCII string 'TRUE'
+// 68		2		Header version
+// 70		2		Required program version
+// 72		4		CRC-32 checksum of the (decrypted) bytes 256-511
+// 76		8		Reserved (set to zero)
+// 84		8		Reserved (set to zero)
+// 92		8		Size of hidden volume in bytes (0 = normal volume)
+// 100		8		Size of the volume in bytes (identical with field 92 for hidden volumes, valid if field 70 >= 0x600 or flag bit 0 == 1)
+// 108		8		Byte offset of the start of the master key scope (valid if field 70 >= 0x600 or flag bit 0 == 1)
+// 116		8		Size of the encrypted area within the master key scope (valid if field 70 >= 0x600 or flag bit 0 == 1)
+// 124		4		Flags: bit 0 set = system encryption; bit 1 set = non-system in-place encryption, bits 2-31 are reserved (set to zero)
+// 128		4		Sector size in bytes
+// 132		120		Reserved (set to zero)
+// 252		4		CRC-32 checksum of the (decrypted) bytes 64-251
+// 256		256		Concatenated primary master key(s) and secondary master key(s) (XTS mode)
 
-/* Volume header v3 structure: */
+
+/* Deprecated/legacy volume header v4 structure (used by TrueCrypt 6.x): */
+//
+// Offset	Length	Description
+// ------------------------------------------
+// Unencrypted:
+// 0		64		Salt
+// Encrypted:
+// 64		4		ASCII string 'TRUE'
+// 68		2		Header version
+// 70		2		Required program version
+// 72		4		CRC-32 checksum of the (decrypted) bytes 256-511
+// 76		8		Reserved (set to zero)
+// 84		8		Reserved (set to zero)
+// 92		8		Size of hidden volume in bytes (0 = normal volume)
+// 100		8		Size of the volume in bytes (identical with field 92 for hidden volumes, valid if field 70 >= 0x600 or flag bit 0 == 1)
+// 108		8		Byte offset of the start of the master key scope (valid if field 70 >= 0x600 or flag bit 0 == 1)
+// 116		8		Size of the encrypted area within the master key scope (valid if field 70 >= 0x600 or flag bit 0 == 1)
+// 124		4		Flags: bit 0 set = system encryption; bit 1 set = non-system in-place encryption, bits 2-31 are reserved
+// 128		124		Reserved (set to zero)
+// 252		4		CRC-32 checksum of the (decrypted) bytes 64-251
+// 256		256		Concatenated primary master key(s) and secondary master key(s) (XTS mode)
+
+
+/* Deprecated/legacy volume header v3 structure (used by TrueCrypt 5.x): */
 //
 // Offset	Length	Description
 // ------------------------------------------
@@ -75,19 +123,19 @@
 
 
 
-uint16 GetHeaderField16 (byte *header, size_t offset)
+uint16 GetHeaderField16 (byte *header, int offset)
 {
 	return BE16 (*(uint16 *) (header + offset));
 }
 
 
-uint32 GetHeaderField32 (byte *header, size_t offset)
+uint32 GetHeaderField32 (byte *header, int offset)
 {
 	return BE32 (*(uint32 *) (header + offset));
 }
 
 
-UINT64_STRUCT GetHeaderField64 (byte *header, size_t offset)
+UINT64_STRUCT GetHeaderField64 (byte *header, int offset)
 {
 	UINT64_STRUCT uint64Struct;
 
@@ -103,23 +151,37 @@ UINT64_STRUCT GetHeaderField64 (byte *header, size_t offset)
 
 #ifndef TC_WINDOWS_BOOT
 
-int VolumeReadHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCRYPTO_INFO *retInfo, CRYPTO_INFO *retHeaderCryptoInfo)
+typedef struct
 {
-	char header[HEADER_SIZE];
+	char DerivedKey[MASTER_KEYDATA_SIZE];
+	BOOL Free;
+	LONG KeyReady;
+	int Pkcs5Prf;
+} KeyDerivationWorkItem;
+
+
+BOOL ReadVolumeHeaderRecoveryMode = FALSE;
+
+int ReadVolumeHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCRYPTO_INFO *retInfo, CRYPTO_INFO *retHeaderCryptoInfo)
+{
+	char header[TC_VOLUME_HEADER_EFFECTIVE_SIZE];
 	KEY_INFO keyInfo;
 	PCRYPTO_INFO cryptoInfo;
 	char dk[MASTER_KEYDATA_SIZE];
-	int pkcs5_prf;
-	int headerVersion, requiredVersion;
-	int status;
+	int enqPkcs5Prf, pkcs5_prf;
+	uint16 headerVersion;
+	int status = ERR_PARAMETER_INCORRECT;
 	int primaryKeyOffset;
 
-#ifdef _WIN32
-#ifndef DEVICE_DRIVER
-	VirtualLock (&keyInfo, sizeof (keyInfo));
-	VirtualLock (&dk, sizeof (dk));
-#endif
-#endif
+	TC_EVENT keyDerivationCompletedEvent;
+	TC_EVENT noOutstandingWorkItemEvent;
+	KeyDerivationWorkItem *keyDerivationWorkItems;
+	KeyDerivationWorkItem *item;
+	int pkcs5PrfCount = LAST_PRF_ID - FIRST_PRF_ID + 1;
+	size_t encryptionThreadCount = GetEncryptionThreadCount();
+	size_t queuedWorkItems = 0;
+	LONG outstandingWorkItemCount = 0;
+	int i;
 
 	if (retHeaderCryptoInfo != NULL)
 	{
@@ -132,46 +194,138 @@ int VolumeReadHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCR
 			return ERR_OUTOFMEMORY;
 	}
 
+	if (encryptionThreadCount > 1)
+	{
+		keyDerivationWorkItems = TCalloc (sizeof (KeyDerivationWorkItem) * pkcs5PrfCount);
+		if (!keyDerivationWorkItems)
+			return ERR_OUTOFMEMORY;
+
+		for (i = 0; i < pkcs5PrfCount; ++i)
+			keyDerivationWorkItems[i].Free = TRUE;
+
+#ifdef DEVICE_DRIVER
+		KeInitializeEvent (&keyDerivationCompletedEvent, SynchronizationEvent, FALSE);
+		KeInitializeEvent (&noOutstandingWorkItemEvent, SynchronizationEvent, TRUE);
+#else
+		keyDerivationCompletedEvent = CreateEvent (NULL, FALSE, FALSE, NULL);
+		if (!keyDerivationCompletedEvent)
+		{
+			TCfree (keyDerivationWorkItems);
+			return ERR_OUTOFMEMORY;
+		}
+
+		noOutstandingWorkItemEvent = CreateEvent (NULL, FALSE, TRUE, NULL);
+		if (!noOutstandingWorkItemEvent)
+		{
+			CloseHandle (keyDerivationCompletedEvent);
+			TCfree (keyDerivationWorkItems);
+			return ERR_OUTOFMEMORY;
+		}
+#endif
+	}
+		
+#ifndef DEVICE_DRIVER
+	VirtualLock (&keyInfo, sizeof (keyInfo));
+	VirtualLock (&dk, sizeof (dk));
+#endif
+
 	crypto_loadkey (&keyInfo, password->Text, (int) password->Length);
 
 	// PKCS5 is used to derive the primary header key(s) and secondary header key(s) (XTS mode) from the password
 	memcpy (keyInfo.salt, encryptedHeader + HEADER_SALT_OFFSET, PKCS5_SALT_SIZE);
 
 	// Test all available PKCS5 PRFs
-	for (pkcs5_prf = FIRST_PRF_ID; pkcs5_prf <= LAST_PRF_ID; pkcs5_prf++)
+	for (enqPkcs5Prf = FIRST_PRF_ID; enqPkcs5Prf <= LAST_PRF_ID || queuedWorkItems > 0; ++enqPkcs5Prf)
 	{
 		BOOL lrw64InitDone = FALSE;		// Deprecated/legacy
 		BOOL lrw128InitDone = FALSE;	// Deprecated/legacy
 
-		keyInfo.noIterations = get_pkcs5_iteration_count (pkcs5_prf, bBoot);
-
-		switch (pkcs5_prf)
+		if (encryptionThreadCount > 1)
 		{
-		case RIPEMD160:
-			derive_key_ripemd160 (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
-				PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
-			break;
+			// Enqueue key derivation on thread pool
+			if (queuedWorkItems < encryptionThreadCount && enqPkcs5Prf <= LAST_PRF_ID)
+			{
+				for (i = 0; i < pkcs5PrfCount; ++i)
+				{
+					item = &keyDerivationWorkItems[i];
+					if (item->Free)
+					{
+						item->Free = FALSE;
+						item->KeyReady = FALSE;
+						item->Pkcs5Prf = enqPkcs5Prf;
 
-		case SHA512:
-			derive_key_sha512 (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
-				PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
-			break;
+						EncryptionThreadPoolBeginKeyDerivation (&keyDerivationCompletedEvent, &noOutstandingWorkItemEvent,
+							&item->KeyReady, &outstandingWorkItemCount, enqPkcs5Prf, keyInfo.userKey,
+							keyInfo.keyLength, keyInfo.salt, get_pkcs5_iteration_count (enqPkcs5Prf, bBoot), item->DerivedKey);
+						
+						++queuedWorkItems;
+						break;
+					}
+				}
 
-		case SHA1:
-			// Deprecated/legacy
-			derive_key_sha1 (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
-				PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
-			break;
+				if (enqPkcs5Prf < LAST_PRF_ID)
+					continue;
+			}
+			else
+				--enqPkcs5Prf;
 
-		case WHIRLPOOL:
-			derive_key_whirlpool (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
-				PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
-			break;
+			// Wait for completion of a key derivation
+			while (queuedWorkItems > 0)
+			{
+				for (i = 0; i < pkcs5PrfCount; ++i)
+				{
+					item = &keyDerivationWorkItems[i];
+					if (!item->Free && InterlockedExchangeAdd (&item->KeyReady, 0) == TRUE)
+					{
+						pkcs5_prf = item->Pkcs5Prf;
+						keyInfo.noIterations = get_pkcs5_iteration_count (pkcs5_prf, bBoot);
+						memcpy (dk, item->DerivedKey, sizeof (dk));
 
-		default:		
-			// Unknown/wrong ID
-			TC_THROW_FATAL_EXCEPTION;
-		} 
+						item->Free = TRUE;
+						--queuedWorkItems;
+						goto KeyReady;
+					}
+				}
+
+				if (queuedWorkItems > 0)
+					TC_WAIT_EVENT (keyDerivationCompletedEvent);
+			}
+			continue;
+KeyReady:	;
+		}
+		else
+		{
+			pkcs5_prf = enqPkcs5Prf;
+			keyInfo.noIterations = get_pkcs5_iteration_count (enqPkcs5Prf, bBoot);
+
+			switch (pkcs5_prf)
+			{
+			case RIPEMD160:
+				derive_key_ripemd160 (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
+					PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
+				break;
+
+			case SHA512:
+				derive_key_sha512 (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
+					PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
+				break;
+
+			case SHA1:
+				// Deprecated/legacy
+				derive_key_sha1 (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
+					PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
+				break;
+
+			case WHIRLPOOL:
+				derive_key_whirlpool (keyInfo.userKey, keyInfo.keyLength, keyInfo.salt,
+					PKCS5_SALT_SIZE, keyInfo.noIterations, dk, GetMaxPkcs5OutSize());
+				break;
+
+			default:		
+				// Unknown/wrong ID
+				TC_THROW_FATAL_EXCEPTION;
+			} 
+		}
 
 		// Test all available modes of operation
 		for (cryptoInfo->mode = FIRST_MODE_OF_OPERATION_ID;
@@ -243,7 +397,7 @@ int VolumeReadHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCR
 				}
 
 				// Copy the header for decryption
-				memcpy (header, encryptedHeader, HEADER_SIZE);
+				memcpy (header, encryptedHeader, sizeof (header));
 
 				// Try to decrypt header 
 
@@ -255,32 +409,52 @@ int VolumeReadHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCR
 
 				// Header version
 				headerVersion = GetHeaderField16 (header, TC_HEADER_OFFSET_VERSION);
-
-				// Required program version
-				requiredVersion = GetHeaderField16 (header, TC_HEADER_OFFSET_REQUIRED_VERSION);
-
-				// Check CRC of the key set
-				if (GetHeaderField32 (header, TC_HEADER_OFFSET_KEY_AREA_CRC) != GetCrc32 (header + HEADER_MASTER_KEYDATA_OFFSET, MASTER_KEYDATA_SIZE))
-					continue;
-
-				// Now we have the correct password, cipher, hash algorithm, and volume type
-
-				// Check the version required to handle this volume
-				if (requiredVersion > VERSION_NUM)
+				
+				if (headerVersion > VOLUME_HEADER_VERSION)
 				{
 					status = ERR_NEW_VERSION_REQUIRED;
 					goto err;
 				}
 
-				// Volume creation time
+				// Check CRC of the header fields
+				if (!ReadVolumeHeaderRecoveryMode
+					&& headerVersion >= 4
+					&& GetHeaderField32 (header, TC_HEADER_OFFSET_HEADER_CRC) != GetCrc32 (header + TC_HEADER_OFFSET_MAGIC, TC_HEADER_OFFSET_HEADER_CRC - TC_HEADER_OFFSET_MAGIC))
+					continue;
+
+				// Required program version
+				cryptoInfo->RequiredProgramVersion = GetHeaderField16 (header, TC_HEADER_OFFSET_REQUIRED_VERSION);
+				cryptoInfo->LegacyVolume = cryptoInfo->RequiredProgramVersion < 0x600;
+
+				// Check CRC of the key set
+				if (!ReadVolumeHeaderRecoveryMode
+					&& GetHeaderField32 (header, TC_HEADER_OFFSET_KEY_AREA_CRC) != GetCrc32 (header + HEADER_MASTER_KEYDATA_OFFSET, MASTER_KEYDATA_SIZE))
+					continue;
+
+				// Now we have the correct password, cipher, hash algorithm, and volume type
+
+				// Check the version required to handle this volume
+				if (cryptoInfo->RequiredProgramVersion > VERSION_NUM)
+				{
+					status = ERR_NEW_VERSION_REQUIRED;
+					goto err;
+				}
+
+				// Header version
+				cryptoInfo->HeaderVersion = headerVersion;
+
+				// Volume creation time (legacy)
 				cryptoInfo->volume_creation_time = GetHeaderField64 (header, TC_HEADER_OFFSET_VOLUME_CREATION_TIME).Value;
 
-				// Header creation time
+				// Header creation time (legacy)
 				cryptoInfo->header_creation_time = GetHeaderField64 (header, TC_HEADER_OFFSET_MODIFICATION_TIME).Value;
 
 				// Hidden volume size (if any)
 				cryptoInfo->hiddenVolumeSize = GetHeaderField64 (header, TC_HEADER_OFFSET_HIDDEN_VOLUME_SIZE).Value;
 
+				// Hidden volume status
+				cryptoInfo->hiddenVolume = (cryptoInfo->hiddenVolumeSize != 0);
+
 				// Volume size
 				cryptoInfo->VolumeSize = GetHeaderField64 (header, TC_HEADER_OFFSET_VOLUME_SIZE);
 				
@@ -288,6 +462,23 @@ int VolumeReadHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCR
 				cryptoInfo->EncryptedAreaStart = GetHeaderField64 (header, TC_HEADER_OFFSET_ENCRYPTED_AREA_START);
 				cryptoInfo->EncryptedAreaLength = GetHeaderField64 (header, TC_HEADER_OFFSET_ENCRYPTED_AREA_LENGTH);
 
+				// Flags
+				cryptoInfo->HeaderFlags = GetHeaderField32 (header, TC_HEADER_OFFSET_FLAGS);
+
+				// Sector size
+				if (headerVersion >= 5)
+					cryptoInfo->SectorSize = GetHeaderField32 (header, TC_HEADER_OFFSET_SECTOR_SIZE);
+				else
+					cryptoInfo->SectorSize = TC_SECTOR_SIZE_LEGACY;
+
+				if (cryptoInfo->SectorSize < TC_MIN_VOLUME_SECTOR_SIZE
+					|| cryptoInfo->SectorSize > TC_MAX_VOLUME_SECTOR_SIZE
+					|| cryptoInfo->SectorSize % ENCRYPTION_DATA_UNIT_SIZE != 0)
+				{
+					status = ERR_PARAMETER_INCORRECT;
+					goto err;
+				}
+
 				// Preserve scheduled header keys if requested			
 				if (retHeaderCryptoInfo)
 				{
@@ -317,7 +508,7 @@ int VolumeReadHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCR
 				cryptoInfo->pkcs5 = pkcs5_prf;
 				cryptoInfo->noIterations = keyInfo.noIterations;
 
-				// Init the encryption algorithm with the decrypted master key
+				// Init the cipher with the decrypted master key
 				status = EAInit (cryptoInfo->ea, keyInfo.master_keydata + primaryKeyOffset, cryptoInfo->ks);
 				if (status == ERR_CIPHER_INIT_FAILURE)
 					goto err;
@@ -346,12 +537,8 @@ int VolumeReadHeader (BOOL bBoot, char *encryptedHeader, Password *password, PCR
 					goto err;
 				}
 
-				// Clear out the temporary key buffers
-ret:
-				burn (dk, sizeof(dk));
-				burn (&keyInfo, sizeof (keyInfo));
-
-				return 0;
+				status = ERR_SUCCESS;
+				goto ret;
 			}
 		}
 	}
@@ -364,14 +551,34 @@ err:
 		*retInfo = NULL; 
 	}
 
+ret:
 	burn (&keyInfo, sizeof (keyInfo));
 	burn (dk, sizeof(dk));
+
+#ifndef DEVICE_DRIVER
+	VirtualUnlock (&keyInfo, sizeof (keyInfo));
+	VirtualUnlock (&dk, sizeof (dk));
+#endif
+
+	if (encryptionThreadCount > 1)
+	{
+		TC_WAIT_EVENT (noOutstandingWorkItemEvent);
+
+		burn (keyDerivationWorkItems, sizeof (KeyDerivationWorkItem) * pkcs5PrfCount);
+		TCfree (keyDerivationWorkItems);
+
+#ifndef DEVICE_DRIVER
+		CloseHandle (keyDerivationCompletedEvent);
+		CloseHandle (noOutstandingWorkItemEvent);
+#endif
+	}
+
 	return status;
 }
 
 #else // TC_WINDOWS_BOOT
 
-int VolumeReadHeader (BOOL bBoot, char *header, Password *password, PCRYPTO_INFO *retInfo, CRYPTO_INFO *retHeaderCryptoInfo)
+int ReadVolumeHeader (BOOL bBoot, char *header, Password *password, PCRYPTO_INFO *retInfo, CRYPTO_INFO *retHeaderCryptoInfo)
 {
 #ifdef TC_WINDOWS_BOOT_SINGLE_CIPHER_MODE
 	char dk[32 * 2];			// 2 * 256-bit key
@@ -391,7 +598,7 @@ int VolumeReadHeader (BOOL bBoot, char *header, Password *password, PCRYPTO_INFO
 
 	// PKCS5 PRF
 	derive_key_ripemd160 (password->Text, (int) password->Length, header + HEADER_SALT_OFFSET,
-		PKCS5_SALT_SIZE, 1000, dk, sizeof (dk));
+		PKCS5_SALT_SIZE, bBoot ? 1000 : 2000, dk, sizeof (dk));
 
 	// Mode of operation
 	cryptoInfo->mode = FIRST_MODE_OF_OPERATION_ID;
@@ -409,8 +616,9 @@ int VolumeReadHeader (BOOL bBoot, char *header, Password *password, PCRYPTO_INFO
 		// Try to decrypt header 
 		DecryptBuffer (header + HEADER_ENCRYPTED_DATA_OFFSET, HEADER_ENCRYPTED_DATA_SIZE, cryptoInfo);
 		
-		// Check magic 'TRUE' and key CRC
+		// Check magic 'TRUE' and CRC-32 of header fields and master keydata
 		if (GetHeaderField32 (header, TC_HEADER_OFFSET_MAGIC) != 0x54525545
+			|| (GetHeaderField16 (header, TC_HEADER_OFFSET_VERSION) >= 4 && GetHeaderField32 (header, TC_HEADER_OFFSET_HEADER_CRC) != GetCrc32 (header + TC_HEADER_OFFSET_MAGIC, TC_HEADER_OFFSET_HEADER_CRC - TC_HEADER_OFFSET_MAGIC))
 			|| GetHeaderField32 (header, TC_HEADER_OFFSET_KEY_AREA_CRC) != GetCrc32 (header + HEADER_MASTER_KEYDATA_OFFSET, MASTER_KEYDATA_SIZE))
 		{
 			EncryptBuffer (header + HEADER_ENCRYPTED_DATA_OFFSET, HEADER_ENCRYPTED_DATA_SIZE, cryptoInfo);
@@ -420,6 +628,10 @@ int VolumeReadHeader (BOOL bBoot, char *header, Password *password, PCRYPTO_INFO
 		// Header decrypted
 		status = 0;
 
+		// Hidden volume status
+		cryptoInfo->VolumeSize = GetHeaderField64 (header, TC_HEADER_OFFSET_HIDDEN_VOLUME_SIZE);
+		cryptoInfo->hiddenVolume = (cryptoInfo->VolumeSize.LowPart != 0 || cryptoInfo->VolumeSize.HighPart != 0);
+
 		// Volume size
 		cryptoInfo->VolumeSize = GetHeaderField64 (header, TC_HEADER_OFFSET_VOLUME_SIZE);
 
@@ -427,6 +639,9 @@ int VolumeReadHeader (BOOL bBoot, char *header, Password *password, PCRYPTO_INFO
 		cryptoInfo->EncryptedAreaStart = GetHeaderField64 (header, TC_HEADER_OFFSET_ENCRYPTED_AREA_START);
 		cryptoInfo->EncryptedAreaLength = GetHeaderField64 (header, TC_HEADER_OFFSET_ENCRYPTED_AREA_LENGTH);
 
+		// Flags
+		cryptoInfo->HeaderFlags = GetHeaderField32 (header, TC_HEADER_OFFSET_FLAGS);
+
 		memcpy (masterKey, header + HEADER_MASTER_KEYDATA_OFFSET, sizeof (masterKey));
 		EncryptBuffer (header + HEADER_ENCRYPTED_DATA_OFFSET, HEADER_ENCRYPTED_DATA_SIZE, cryptoInfo);
 
@@ -464,14 +679,15 @@ ret:
 #if !defined (DEVICE_DRIVER) && !defined (TC_WINDOWS_BOOT)
 
 #ifdef VOLFORMAT
-#include "../Format/TcFormat.h"
+#	include "../Format/TcFormat.h"
+#	include "Dlgcode.h"
 #endif
 
 // Creates a volume header in memory
-int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *password,
-		   int pkcs5_prf, char *masterKeydata, unsigned __int64 volumeCreationTime, PCRYPTO_INFO *retInfo,
+int CreateVolumeHeaderInMemory (BOOL bBoot, char *header, int ea, int mode, Password *password,
+		   int pkcs5_prf, char *masterKeydata, PCRYPTO_INFO *retInfo,
 		   unsigned __int64 volumeSize, unsigned __int64 hiddenVolumeSize,
-		   unsigned __int64 encryptedAreaStart, unsigned __int64 encryptedAreaLength, BOOL bWipeMode)
+		   unsigned __int64 encryptedAreaStart, unsigned __int64 encryptedAreaLength, uint16 requiredProgramVersion, uint32 headerFlags, uint32 sectorSize, BOOL bWipeMode)
 {
 	unsigned char *p = (unsigned char *) header;
 	static KEY_INFO keyInfo;
@@ -486,12 +702,10 @@ int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *pas
 	if (cryptoInfo == NULL)
 		return ERR_OUTOFMEMORY;
 
-	memset (header, 0, HEADER_SIZE);
+	memset (header, 0, TC_VOLUME_HEADER_EFFECTIVE_SIZE);
 
-#ifdef _WIN32
 	VirtualLock (&keyInfo, sizeof (keyInfo));
 	VirtualLock (&dk, sizeof (dk));
-#endif
 
 	/* Encryption setup */
 
@@ -511,9 +725,10 @@ int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *pas
 			// Deprecated/legacy modes of operation
 			bytesNeeded = LEGACY_VOL_IV_SIZE + EAGetKeySize (ea);
 
-			/* In fact, this should never be the case since new volumes are not supposed to use
-			   any deprecated mode of operation. */
-			return ERR_VOL_FORMAT_BAD;
+			// In fact, this should never be the case since volumes being newly created are not
+			// supposed to use any deprecated mode of operation.
+			TC_THROW_FATAL_EXCEPTION;
+			break;
 
 		default:
 			bytesNeeded = EAGetKeySize (ea) * 2;	// Size of primary + secondary key(s)
@@ -581,18 +796,8 @@ int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *pas
 	mputLong (p, 0x54525545);
 
 	// Header version
-	switch (mode)
-	{
-	case LRW:
-	case CBC:
-	case OUTER_CBC:
-	case INNER_CBC:
-		// Deprecated/legacy modes (used before TrueCrypt 5.0)
-		mputWord (p, 0x0002);
-		break;
-	default:
-		mputWord (p, VOLUME_HEADER_VERSION);
-	}
+	mputWord (p, VOLUME_HEADER_VERSION);
+	cryptoInfo->HeaderVersion = VOLUME_HEADER_VERSION;
 
 	// Required program version to handle this volume
 	switch (mode)
@@ -611,57 +816,15 @@ int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *pas
 		mputWord (p, hiddenVolumeSize > 0 ? 0x0300 : 0x0100);
 		break;
 	default:
-		mputWord (p, VOL_REQ_PROG_VERSION);
+		mputWord (p, requiredProgramVersion != 0 ? requiredProgramVersion : TC_VOLUME_MIN_REQUIRED_PROGRAM_VERSION);
 	}
 
 	// CRC of the master key data
 	x = GetCrc32(keyInfo.master_keydata, MASTER_KEYDATA_SIZE);
 	mputLong (p, x);
 
-	// Time
-	{
-#ifdef _WIN32
-		SYSTEMTIME st;
-		FILETIME ft;
-
-		// Volume creation time
-		if (volumeCreationTime == 0)
-		{
-			GetLocalTime (&st);
-			SystemTimeToFileTime (&st, &ft);
-		}
-		else
-		{
-			ft.dwHighDateTime = (DWORD)(volumeCreationTime >> 32);
-			ft.dwLowDateTime = (DWORD)volumeCreationTime;
-		}
-		mputLong (p, ft.dwHighDateTime);
-		mputLong (p, ft.dwLowDateTime);
-
-		// Header modification time/date
-		GetLocalTime (&st);
-		SystemTimeToFileTime (&st, &ft);
-		mputLong (p, ft.dwHighDateTime);
-		mputLong (p, ft.dwLowDateTime);
-
-#else
-		struct timeval tv;
-		unsigned __int64 ct, wt;
-		gettimeofday (&tv, NULL);
-
-		// Unix time => Windows file time
-		wt = ((unsigned __int64)tv.tv_sec + 134774LL * 24 * 3600) * 1000LL * 1000 * 10;
-
-		if (volumeCreationTime == 0)
-			ct = wt;
-		else
-			ct = volumeCreationTime;
-
-		mputInt64 (p, ct);
-		mputInt64 (p, wt);
-#endif
-
-	}
+	// Reserved fields
+	p += 2 * 8;
 
 	// Size of hidden volume (if any)
 	cryptoInfo->hiddenVolumeSize = hiddenVolumeSize;
@@ -681,6 +844,26 @@ int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *pas
 	cryptoInfo->EncryptedAreaLength.Value = encryptedAreaLength;
 	mputInt64 (p, encryptedAreaLength);
 
+	// Flags
+	cryptoInfo->HeaderFlags = headerFlags;
+	mputLong (p, headerFlags);
+
+	// Sector size
+	if (sectorSize < TC_MIN_VOLUME_SECTOR_SIZE
+		|| sectorSize > TC_MAX_VOLUME_SECTOR_SIZE
+		|| sectorSize % ENCRYPTION_DATA_UNIT_SIZE != 0)
+	{
+		TC_THROW_FATAL_EXCEPTION;
+	}
+
+	cryptoInfo->SectorSize = sectorSize;
+	mputLong (p, sectorSize);
+
+	// CRC of the header fields
+	x = GetCrc32 (header + TC_HEADER_OFFSET_MAGIC, TC_HEADER_OFFSET_HEADER_CRC - TC_HEADER_OFFSET_MAGIC);
+	p = header + TC_HEADER_OFFSET_HEADER_CRC;
+	mputLong (p, x);
+
 	// The master key data
 	memcpy (header + HEADER_MASTER_KEYDATA_OFFSET, keyInfo.master_keydata, MASTER_KEYDATA_SIZE);
 
@@ -753,7 +936,7 @@ int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *pas
 
 
 #ifdef VOLFORMAT
-	if (showKeys)
+	if (showKeys && !bInPlaceEncNonSys)
 	{
 		BOOL dots3 = FALSE;
 		int i, j;
@@ -805,4 +988,217 @@ int VolumeWriteHeader (BOOL bBoot, char *header, int ea, int mode, Password *pas
 	return 0;
 }
 
+
+BOOL ReadEffectiveVolumeHeader (BOOL device, HANDLE fileHandle, byte *header, DWORD *bytesRead)
+{
+#if TC_VOLUME_HEADER_EFFECTIVE_SIZE > TC_MAX_VOLUME_SECTOR_SIZE
+#error TC_VOLUME_HEADER_EFFECTIVE_SIZE > TC_MAX_VOLUME_SECTOR_SIZE
+#endif
+
+	byte sectorBuffer[TC_MAX_VOLUME_SECTOR_SIZE];
+	DISK_GEOMETRY geometry;
+
+	if (!device)
+		return ReadFile (fileHandle, header, TC_VOLUME_HEADER_EFFECTIVE_SIZE, bytesRead, NULL);
+
+	if (!DeviceIoControl (fileHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof (geometry), bytesRead, NULL))
+		return FALSE;
+
+	if (geometry.BytesPerSector > sizeof (sectorBuffer) || geometry.BytesPerSector < TC_MIN_VOLUME_SECTOR_SIZE)
+	{
+		SetLastError (ERROR_INVALID_PARAMETER);
+		return FALSE;
+	}
+
+	if (!ReadFile (fileHandle, sectorBuffer, max (TC_VOLUME_HEADER_EFFECTIVE_SIZE, geometry.BytesPerSector), bytesRead, NULL))
+		return FALSE;
+
+	memcpy (header, sectorBuffer, min (*bytesRead, TC_VOLUME_HEADER_EFFECTIVE_SIZE));
+	
+	if (*bytesRead > TC_VOLUME_HEADER_EFFECTIVE_SIZE)
+		*bytesRead = TC_VOLUME_HEADER_EFFECTIVE_SIZE;
+
+	return TRUE;
+}
+
+
+BOOL WriteEffectiveVolumeHeader (BOOL device, HANDLE fileHandle, byte *header)
+{
+#if TC_VOLUME_HEADER_EFFECTIVE_SIZE > TC_MAX_VOLUME_SECTOR_SIZE
+#error TC_VOLUME_HEADER_EFFECTIVE_SIZE > TC_MAX_VOLUME_SECTOR_SIZE
+#endif
+
+	byte sectorBuffer[TC_MAX_VOLUME_SECTOR_SIZE];
+	DWORD bytesDone;
+	DISK_GEOMETRY geometry;
+
+	if (!device)
+	{
+		if (!WriteFile (fileHandle, header, TC_VOLUME_HEADER_EFFECTIVE_SIZE, &bytesDone, NULL))
+			return FALSE;
+
+		if (bytesDone != TC_VOLUME_HEADER_EFFECTIVE_SIZE)
+		{
+			SetLastError (ERROR_INVALID_PARAMETER);
+			return FALSE;
+		}
+
+		return TRUE;
+	}
+
+	if (!DeviceIoControl (fileHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof (geometry), &bytesDone, NULL))
+		return FALSE;
+
+	if (geometry.BytesPerSector > sizeof (sectorBuffer) || geometry.BytesPerSector < TC_MIN_VOLUME_SECTOR_SIZE)
+	{
+		SetLastError (ERROR_INVALID_PARAMETER);
+		return FALSE;
+	}
+
+	if (geometry.BytesPerSector != TC_VOLUME_HEADER_EFFECTIVE_SIZE)
+	{
+		LARGE_INTEGER seekOffset;
+
+		if (!ReadFile (fileHandle, sectorBuffer, geometry.BytesPerSector, &bytesDone, NULL))
+			return FALSE;
+
+		if (bytesDone != geometry.BytesPerSector)
+		{
+			SetLastError (ERROR_INVALID_PARAMETER);
+			return FALSE;
+		}
+
+		seekOffset.QuadPart = -(int) bytesDone;
+		if (!SetFilePointerEx (fileHandle, seekOffset, NULL, FILE_CURRENT))
+			return FALSE;
+	}
+
+	memcpy (sectorBuffer, header, TC_VOLUME_HEADER_EFFECTIVE_SIZE);
+
+	if (!WriteFile (fileHandle, sectorBuffer, geometry.BytesPerSector, &bytesDone, NULL))
+		return FALSE;
+
+	if (bytesDone != geometry.BytesPerSector)
+	{
+		SetLastError (ERROR_INVALID_PARAMETER);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+
+// Writes randomly generated data to unused/reserved header areas.
+// When bPrimaryOnly is TRUE, then only the primary header area (not the backup header area) is filled with random data.
+// When bBackupOnly is TRUE, only the backup header area (not the primary header area) is filled with random data.
+int WriteRandomDataToReservedHeaderAreas (HANDLE dev, CRYPTO_INFO *cryptoInfo, uint64 dataAreaSize, BOOL bPrimaryOnly, BOOL bBackupOnly)
+{
+	char temporaryKey[MASTER_KEYDATA_SIZE];
+	char originalK2[MASTER_KEYDATA_SIZE];
+
+	byte buf[TC_VOLUME_HEADER_GROUP_SIZE];
+
+	LARGE_INTEGER offset;
+	int nStatus = ERR_SUCCESS;
+	DWORD dwError;
+	DWORD bytesDone;
+	BOOL backupHeaders = bBackupOnly;
+
+	if (bPrimaryOnly && bBackupOnly)
+		TC_THROW_FATAL_EXCEPTION;
+
+	memcpy (originalK2, cryptoInfo->k2, sizeof (cryptoInfo->k2));
+
+	while (TRUE)
+	{
+		// Temporary keys
+		if (!RandgetBytes (temporaryKey, EAGetKeySize (cryptoInfo->ea), FALSE)
+			|| !RandgetBytes (cryptoInfo->k2, sizeof (cryptoInfo->k2), FALSE))
+		{
+			nStatus = ERR_PARAMETER_INCORRECT; 
+			goto final_seq;
+		}
+
+		nStatus = EAInit (cryptoInfo->ea, temporaryKey, cryptoInfo->ks);
+		if (nStatus != ERR_SUCCESS)
+			goto final_seq;
+
+		if (!EAInitMode (cryptoInfo))
+		{
+			nStatus = ERR_MODE_INIT_FAILED;
+			goto final_seq;
+		}
+
+		offset.QuadPart = backupHeaders ? dataAreaSize + TC_VOLUME_HEADER_GROUP_SIZE : TC_VOLUME_HEADER_OFFSET;
+
+		if (!SetFilePointerEx (dev, offset, NULL, FILE_BEGIN))
+		{
+			nStatus = ERR_OS_ERROR;
+			goto final_seq;
+		}
+
+		if (!ReadFile (dev, buf, sizeof (buf), &bytesDone, NULL))
+		{
+			nStatus = ERR_OS_ERROR;
+			goto final_seq;
+		}
+
+		if (bytesDone < TC_VOLUME_HEADER_EFFECTIVE_SIZE)
+		{
+			SetLastError (ERROR_INVALID_PARAMETER);
+			nStatus = ERR_OS_ERROR;
+			goto final_seq;
+		}
+
+		EncryptBuffer (buf + TC_VOLUME_HEADER_EFFECTIVE_SIZE, sizeof (buf) - TC_VOLUME_HEADER_EFFECTIVE_SIZE, cryptoInfo);
+
+		if (!SetFilePointerEx (dev, offset, NULL, FILE_BEGIN))
+		{
+			nStatus = ERR_OS_ERROR;
+			goto final_seq;
+		}
+
+		if (!WriteFile (dev, buf, sizeof (buf), &bytesDone, NULL))
+		{
+			nStatus = ERR_OS_ERROR;
+			goto final_seq;
+		}
+
+		if (bytesDone != sizeof (buf))
+		{
+			nStatus = ERR_PARAMETER_INCORRECT;
+			goto final_seq;
+		}
+
+		if (backupHeaders || bPrimaryOnly)
+			break;
+
+		backupHeaders = TRUE;
+	}
+
+	memcpy (cryptoInfo->k2, originalK2, sizeof (cryptoInfo->k2));
+
+	nStatus = EAInit (cryptoInfo->ea, cryptoInfo->master_keydata, cryptoInfo->ks);
+	if (nStatus != ERR_SUCCESS)
+		goto final_seq;
+
+	if (!EAInitMode (cryptoInfo))
+	{
+		nStatus = ERR_MODE_INIT_FAILED;
+		goto final_seq;
+	}
+
+final_seq:
+
+	dwError = GetLastError();
+
+	burn (temporaryKey, sizeof (temporaryKey));
+	burn (originalK2, sizeof (originalK2));
+
+	if (nStatus != ERR_SUCCESS)
+		SetLastError (dwError);
+
+	return nStatus;
+}
+
 #endif // !defined (DEVICE_DRIVER) && !defined (TC_WINDOWS_BOOT)
