@@ -1,15 +1,149 @@
 /*
- Copyright (c) 2008 TrueCrypt Foundation. All rights reserved.
+ Copyright (c) 2008-2009 TrueCrypt Developers Association. All rights reserved.
 
- Governed by the TrueCrypt License 2.4 the full text of which is contained
- in the file License.txt included in TrueCrypt binary and source code
- distribution packages.
+ Governed by the TrueCrypt License 3.0 the full text of which is contained in
+ the file License.txt included in TrueCrypt binary and source code distribution
+ packages.
 */
 
 #include "TCdefs.h"
 #include "Apidrvr.h"
 #include "Ntdriver.h"
+#include "DriveFilter.h"
 #include "EncryptedIoQueue.h"
+#include "EncryptionThreadPool.h"
+#include "Volumes.h"
+
+
+static void AcquireBufferPoolMutex (EncryptedIoQueue *queue)
+{
+	NTSTATUS status;
+
+	status = KeWaitForMutexObject (&queue->BufferPoolMutex, Executive, KernelMode, FALSE, NULL);
+	if (!NT_SUCCESS (status))
+		TC_BUG_CHECK (status);
+}
+
+
+static void ReleaseBufferPoolMutex (EncryptedIoQueue *queue)
+{
+	KeReleaseMutex (&queue->BufferPoolMutex, FALSE);
+}
+
+
+static void *GetPoolBuffer (EncryptedIoQueue *queue, ULONG requestedSize)
+{
+	EncryptedIoQueueBuffer *buffer;
+	void *bufferAddress = NULL;
+	BOOL requestedSizePresentInPool = FALSE;
+
+	while (TRUE)
+	{
+		AcquireBufferPoolMutex (queue);
+
+		for (buffer = queue->FirstPoolBuffer; ; buffer = buffer->NextBuffer)
+		{
+			if (buffer && buffer->Size == requestedSize)
+			{
+				requestedSizePresentInPool = TRUE;
+
+				if (!buffer->InUse)
+				{
+					// Reuse a free buffer
+					buffer->InUse = TRUE;
+					bufferAddress = buffer->Address;
+					break;
+				}
+			}
+
+			if (!buffer || !buffer->NextBuffer)
+			{
+				EncryptedIoQueueBuffer *newBuffer;
+
+				if (requestedSizePresentInPool && !queue->StartPending)
+					break;
+
+				// Allocate a new buffer
+				newBuffer = TCalloc (sizeof (EncryptedIoQueueBuffer));
+				if (!newBuffer)
+				{
+					bufferAddress = NULL;
+					break;
+				}
+
+				bufferAddress = TCalloc (requestedSize);
+				if (bufferAddress)
+				{
+					newBuffer->NextBuffer = NULL;
+					newBuffer->Address = bufferAddress;
+					newBuffer->Size = requestedSize;
+					newBuffer->InUse = TRUE;
+
+					if (!buffer)
+						queue->FirstPoolBuffer = newBuffer;
+					else
+						buffer->NextBuffer = newBuffer;
+				}
+				else
+					TCfree (newBuffer);
+
+				break;
+			}
+		}
+
+		ReleaseBufferPoolMutex (queue);
+
+		if (bufferAddress || !requestedSizePresentInPool || queue->StartPending)
+			break;
+
+		KeWaitForSingleObject (&queue->PoolBufferFreeEvent, Executive, KernelMode, FALSE, NULL);
+	}
+
+	return bufferAddress;
+}
+
+
+static void ReleasePoolBuffer (EncryptedIoQueue *queue, void *address)
+{
+	EncryptedIoQueueBuffer *buffer;
+	AcquireBufferPoolMutex (queue);
+	
+	for (buffer = queue->FirstPoolBuffer; buffer != NULL; buffer = buffer->NextBuffer)
+	{
+		if (buffer->Address == address)
+		{
+			ASSERT (buffer->InUse);
+
+			buffer->InUse = FALSE;
+			break;
+		}
+	}
+
+	ReleaseBufferPoolMutex (queue);
+	KeSetEvent (&queue->PoolBufferFreeEvent, IO_DISK_INCREMENT, FALSE);
+}
+
+
+static void FreePoolBuffers (EncryptedIoQueue *queue)
+{
+	EncryptedIoQueueBuffer *buffer;
+	AcquireBufferPoolMutex (queue);
+
+	for (buffer = queue->FirstPoolBuffer; buffer != NULL; )
+	{
+		EncryptedIoQueueBuffer *nextBuffer = buffer->NextBuffer;
+
+		ASSERT (!buffer->InUse);
+
+		TCfree (buffer->Address);
+		TCfree (buffer);
+
+		buffer = nextBuffer;
+	}
+
+	queue->FirstPoolBuffer = NULL;
+	ReleaseBufferPoolMutex (queue);
+}
 
 
 static void DecrementOutstandingIoCount (EncryptedIoQueue *queue)
@@ -19,12 +153,10 @@ static void DecrementOutstandingIoCount (EncryptedIoQueue *queue)
 }
 
 
-static void OnItemCompleted (EncryptedIoQueueItem *item)
+static void OnItemCompleted (EncryptedIoQueueItem *item, BOOL freeItem)
 {
 	DecrementOutstandingIoCount (item->Queue);
-	
-	if (item->Queue->IsFilterDevice)
-		IoReleaseRemoveLock (&item->Queue->RemoveLock, item->OriginalIrp);
+	IoReleaseRemoveLock (&item->Queue->RemoveLock, item->OriginalIrp);
 
 	if (NT_SUCCESS (item->Status))
 	{
@@ -34,15 +166,22 @@ static void OnItemCompleted (EncryptedIoQueueItem *item)
 			item->Queue->TotalBytesRead += item->OriginalLength;
 	}
 
-	TCfree (item);
+	if (freeItem)
+		ReleasePoolBuffer (item->Queue, item);
 }
 
 
 static NTSTATUS CompleteOriginalIrp (EncryptedIoQueueItem *item, NTSTATUS status, ULONG_PTR information)
 {
-	//Dump ("Queue comp  offset=%I64d  status=%x  info=%p  out=%d\n", item->OriginalOffset, status, information, item->Queue->OutstandingIoCount - 1);
+#ifdef TC_TRACE_IO_QUEUE
+	Dump ("< %I64d [%I64d] %c status=%x info=%I64d\n", item->OriginalIrpOffset, GetElapsedTime (&item->Queue->LastPerformanceCounter), item->Write ? 'W' : 'R', status, (int64) information);
+#endif
+
 	TCCompleteDiskIrp (item->OriginalIrp, status, information);
-	OnItemCompleted (item);
+
+	item->Status = status;
+	OnItemCompleted (item, TRUE);
+
 	return status;
 }
 
@@ -89,6 +228,9 @@ static VOID CompletionThreadProc (PVOID threadArg)
 	EncryptedIoRequest *request;
 	UINT64_STRUCT dataUnit;
 
+	if (IsEncryptionThreadPoolRunning())
+		KeSetPriorityThread (KeGetCurrentThread(), LOW_REALTIME_PRIORITY);
+
 	while (!queue->ThreadExitRequested)
 	{
 		if (!NT_SUCCESS (KeWaitForSingleObject (&queue->CompletionThreadQueueNotEmptyEvent, Executive, KernelMode, FALSE, NULL)))
@@ -108,6 +250,8 @@ static VOID CompletionThreadProc (PVOID threadArg)
 
 				if (queue->CryptoInfo->bPartitionInInactiveSysEncScope)
 					dataUnit.Value += queue->CryptoInfo->FirstDataUnitNo.Value;
+				else if (queue->RemapEncryptedArea)
+					dataUnit.Value += queue->RemappedAreaDataUnitOffset;
 
 				DecryptDataUnits (request->Data + request->EncryptedOffset, &dataUnit, request->EncryptedLength / ENCRYPTION_DATA_UNIT_SIZE, queue->CryptoInfo);
 			}
@@ -117,13 +261,8 @@ static VOID CompletionThreadProc (PVOID threadArg)
 				CompleteOriginalIrp (request->Item, request->Item->Status,
 					NT_SUCCESS (request->Item->Status) ? request->Item->OriginalLength : 0);
 			}
-			else
-			{
-				InterlockedDecrement (&request->Item->OutstandingRequestCount);
-				KeSetEvent (&queue->RequestCompletedEvent, IO_DISK_INCREMENT, FALSE);
-			}
 
-			TCfree (request);
+			ReleasePoolBuffer (queue, request);
 		}
 	}
 
@@ -131,6 +270,31 @@ static VOID CompletionThreadProc (PVOID threadArg)
 }
 
 
+static NTSTATUS TCCachedRead (EncryptedIoQueue *queue, IO_STATUS_BLOCK *ioStatus, PVOID buffer, LARGE_INTEGER offset, ULONG length)
+{
+	queue->LastReadOffset = offset;
+	queue->LastReadLength = length;
+
+	if (queue->ReadAheadBufferValid && queue->ReadAheadOffset.QuadPart == offset.QuadPart && queue->ReadAheadLength >= length)
+	{
+		memcpy (buffer, queue->ReadAheadBuffer, length);
+
+		if (!queue->IsFilterDevice)
+		{
+			ioStatus->Information = length;
+			ioStatus->Status = STATUS_SUCCESS;
+		}
+
+		return STATUS_SUCCESS;
+	}
+
+	if (queue->IsFilterDevice)
+		return TCReadDevice (queue->LowerDeviceObject, buffer, offset, length);
+
+	return ZwReadFile (queue->HostFileHandle, NULL, NULL, NULL, ioStatus, buffer, length, &offset, NULL);
+}
+
+
 static VOID IoThreadProc (PVOID threadArg)
 {
 	EncryptedIoQueue *queue = (EncryptedIoQueue *) threadArg;
@@ -139,6 +303,15 @@ static VOID IoThreadProc (PVOID threadArg)
 
 	KeSetPriorityThread (KeGetCurrentThread(), LOW_REALTIME_PRIORITY);
 
+	if (!queue->IsFilterDevice && queue->SecurityClientContext)
+	{
+#ifdef DEBUG
+		NTSTATUS status =
+#endif
+		SeImpersonateClientEx (queue->SecurityClientContext, NULL);
+		ASSERT (NT_SUCCESS (status));
+	}
+
 	while (!queue->ThreadExitRequested)
 	{
 		if (!NT_SUCCESS (KeWaitForSingleObject (&queue->IoThreadQueueNotEmptyEvent, Executive, KernelMode, FALSE, NULL)))
@@ -149,17 +322,79 @@ static VOID IoThreadProc (PVOID threadArg)
 
 		while ((listEntry = ExInterlockedRemoveHeadList (&queue->IoThreadQueue, &queue->IoThreadQueueLock)))
 		{
+			InterlockedDecrement (&queue->IoThreadPendingRequestCount);
 			request = CONTAINING_RECORD (listEntry, EncryptedIoRequest, ListEntry);
 			
+#ifdef TC_TRACE_IO_QUEUE
+			Dump ("%c   %I64d [%I64d] roff=%I64d rlen=%d\n", request->Item->Write ? 'W' : 'R', request->Item->OriginalIrpOffset.QuadPart, GetElapsedTime (&queue->LastPerformanceCounter), request->Offset.QuadPart, request->Length);
+#endif
+
 			// Perform IO request if no preceding request of the item failed
 			if (NT_SUCCESS (request->Item->Status))
 			{
 				if (queue->IsFilterDevice)
 				{
-					if (request->Item->Write)
-						request->Item->Status = TCWriteDevice (queue->LowerDeviceObject, request->Data, request->Offset, request->Length);
+					if (queue->RemapEncryptedArea && request->EncryptedLength > 0)
+					{
+						if (request->EncryptedLength != request->Length)
+						{
+							// Up to three subfragments may be required to handle a partially remapped fragment
+							int subFragment;
+							byte *subFragmentData = request->Data;
+
+							for (subFragment = 0 ; subFragment < 3; ++subFragment)
+							{
+								LARGE_INTEGER subFragmentOffset;
+								ULONG subFragmentLength;
+								subFragmentOffset.QuadPart = request->Offset.QuadPart;
+
+								switch (subFragment)
+								{
+								case 0:
+									subFragmentLength = (ULONG) request->EncryptedOffset;
+									break;
+
+								case 1:
+									subFragmentOffset.QuadPart += request->EncryptedOffset + queue->RemappedAreaOffset;
+									subFragmentLength = request->EncryptedLength;
+									break;
+
+								case 2:
+									subFragmentOffset.QuadPart += request->EncryptedOffset + request->EncryptedLength;
+									subFragmentLength = (ULONG) (request->Length - (request->EncryptedOffset + request->EncryptedLength));
+									break;
+								}
+
+								if (subFragmentLength > 0)
+								{
+									if (request->Item->Write)
+										request->Item->Status = TCWriteDevice (queue->LowerDeviceObject, subFragmentData, subFragmentOffset, subFragmentLength);
+									else
+										request->Item->Status = TCCachedRead (queue, NULL, subFragmentData, subFragmentOffset, subFragmentLength);
+
+									subFragmentData += subFragmentLength;
+								}
+							}
+						}
+						else
+						{
+							// Remap the fragment
+							LARGE_INTEGER remappedOffset;
+							remappedOffset.QuadPart = request->Offset.QuadPart + queue->RemappedAreaOffset;
+
+							if (request->Item->Write)
+								request->Item->Status = TCWriteDevice (queue->LowerDeviceObject, request->Data, remappedOffset, request->Length);
+							else
+								request->Item->Status = TCCachedRead (queue, NULL, request->Data, remappedOffset, request->Length);
+						}
+					}
 					else
-						request->Item->Status = TCReadDevice (queue->LowerDeviceObject, request->Data, request->Offset, request->Length);
+					{
+						if (request->Item->Write)
+							request->Item->Status = TCWriteDevice (queue->LowerDeviceObject, request->Data, request->Offset, request->Length);
+						else
+							request->Item->Status = TCCachedRead (queue, NULL, request->Data, request->Offset, request->Length);
+					}
 				}
 				else
 				{
@@ -168,12 +403,17 @@ static VOID IoThreadProc (PVOID threadArg)
 					if (request->Item->Write)
 						request->Item->Status = ZwWriteFile (queue->HostFileHandle, NULL, NULL, NULL, &ioStatus, request->Data, request->Length, &request->Offset, NULL);
 					else
-						request->Item->Status = ZwReadFile (queue->HostFileHandle, NULL, NULL, NULL, &ioStatus, request->Data, request->Length, &request->Offset, NULL);
+						request->Item->Status = TCCachedRead (queue, &ioStatus, request->Data, request->Offset, request->Length);
+
+					if (NT_SUCCESS (request->Item->Status) && ioStatus.Information != request->Length)
+						request->Item->Status = STATUS_END_OF_FILE;
 				}
 			}
 
 			if (request->Item->Write)
 			{
+				queue->ReadAheadBufferValid = FALSE;
+
 				ReleaseFragmentBuffer (queue, request->Data);
 
 				if (request->CompleteOriginalIrp)
@@ -181,27 +421,56 @@ static VOID IoThreadProc (PVOID threadArg)
 					CompleteOriginalIrp (request->Item, request->Item->Status,
 						NT_SUCCESS (request->Item->Status) ? request->Item->OriginalLength : 0);
 				}
-				else
-				{
-					InterlockedDecrement (&request->Item->OutstandingRequestCount);
-					KeSetEvent (&queue->RequestCompletedEvent, IO_DISK_INCREMENT, FALSE);
-				}
 
-				TCfree (request);
+				ReleasePoolBuffer (queue, request);
 			}
 			else
 			{
+				BOOL readAhead = FALSE;
+
 				if (NT_SUCCESS (request->Item->Status))
-				{
-					// Copy fragment to original IRP buffer
 					memcpy (request->OrigDataBufferFragment, request->Data, request->Length);
-				}
 
 				ReleaseFragmentBuffer (queue, request->Data);
 				request->Data = request->OrigDataBufferFragment;
 
+				if (request->CompleteOriginalIrp
+					&& queue->LastReadLength > 0
+					&& NT_SUCCESS (request->Item->Status)
+					&& InterlockedExchangeAdd (&queue->IoThreadPendingRequestCount, 0) == 0)
+				{
+					readAhead = TRUE;
+					InterlockedIncrement (&queue->OutstandingIoCount);
+				}
+
 				ExInterlockedInsertTailList (&queue->CompletionThreadQueue, &request->CompletionListEntry, &queue->CompletionThreadQueueLock);
 				KeSetEvent (&queue->CompletionThreadQueueNotEmptyEvent, IO_DISK_INCREMENT, FALSE);
+
+				if (readAhead)
+				{
+					queue->ReadAheadBufferValid = FALSE;
+					queue->ReadAheadOffset.QuadPart = queue->LastReadOffset.QuadPart + queue->LastReadLength;
+					queue->ReadAheadLength = queue->LastReadLength;
+
+					if (queue->ReadAheadOffset.QuadPart + queue->ReadAheadLength <= queue->MaxReadAheadOffset.QuadPart)
+					{
+#ifdef TC_TRACE_IO_QUEUE
+						Dump ("A   %I64d [%I64d] roff=%I64d rlen=%d\n", request->Item->OriginalIrpOffset.QuadPart, GetElapsedTime (&queue->LastPerformanceCounter), queue->ReadAheadOffset, queue->ReadAheadLength);
+#endif
+						if (queue->IsFilterDevice)
+						{
+							queue->ReadAheadBufferValid = NT_SUCCESS (TCReadDevice (queue->LowerDeviceObject, queue->ReadAheadBuffer, queue->ReadAheadOffset, queue->ReadAheadLength));
+						}
+						else
+						{
+							IO_STATUS_BLOCK ioStatus;
+							queue->ReadAheadBufferValid = NT_SUCCESS (ZwReadFile (queue->HostFileHandle, NULL, NULL, NULL, &ioStatus, queue->ReadAheadBuffer, queue->ReadAheadLength, &queue->ReadAheadOffset, NULL));
+							queue->ReadAheadLength = (ULONG) ioStatus.Information;
+						}
+					}
+
+					DecrementOutstandingIoCount (queue);
+				}
 			}
 		}
 	}
@@ -210,16 +479,6 @@ static VOID IoThreadProc (PVOID threadArg)
 }
 
 
-static NTSTATUS OnPassedIrpCompleted (PDEVICE_OBJECT filterDeviceObject, PIRP irp, EncryptedIoQueueItem *item)
-{
-	if (irp->PendingReturned)
-		IoMarkIrpPending (irp);
-
-	OnItemCompleted (item);
-	return STATUS_CONTINUE_COMPLETION;
-}
-
-
 static VOID MainThreadProc (PVOID threadArg)
 {
 	EncryptedIoQueue *queue = (EncryptedIoQueue *) threadArg;
@@ -234,6 +493,9 @@ static VOID MainThreadProc (PVOID threadArg)
 	uint64 intersectStart;
 	uint32 intersectLength;
 
+	if (IsEncryptionThreadPoolRunning())
+		KeSetPriorityThread (KeGetCurrentThread(), LOW_REALTIME_PRIORITY);
+
 	while (!queue->ThreadExitRequested)
 	{
 		if (!NT_SUCCESS (KeWaitForSingleObject (&queue->MainThreadQueueNotEmptyEvent, Executive, KernelMode, FALSE, NULL)))
@@ -241,26 +503,95 @@ static VOID MainThreadProc (PVOID threadArg)
 
 		while ((listEntry = ExInterlockedRemoveHeadList (&queue->MainThreadQueue, &queue->MainThreadQueueLock)))
 		{
-			item = CONTAINING_RECORD (listEntry, EncryptedIoQueueItem, ListEntry);
-
-			if (queue->Suspended)
-			{
-				KeWaitForSingleObject (&queue->QueueResumedEvent, Executive, KernelMode, FALSE, NULL);
-			}
+			PIRP irp = CONTAINING_RECORD (listEntry, IRP, Tail.Overlay.ListEntry);
+			PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation (irp);
 			
-			IoSetCancelRoutine (item->OriginalIrp, NULL);
-			if (item->OriginalIrp->Cancel)
+			if (queue->Suspended)
+				KeWaitForSingleObject (&queue->QueueResumedEvent, Executive, KernelMode, FALSE, NULL);
+
+			item = GetPoolBuffer (queue, sizeof (EncryptedIoQueueItem));
+			item->Queue = queue;
+			item->OriginalIrp = irp;
+			item->Status = STATUS_SUCCESS;
+
+			IoSetCancelRoutine (irp, NULL);
+			if (irp->Cancel)
 			{
 				CompleteOriginalIrp (item, STATUS_CANCELLED, 0);
 				continue;
 			}
 
-			// Pass the IRP if the drive is not encrypted
-			if (queue->IsFilterDevice && (queue->EncryptedAreaStart == -1 || queue->EncryptedAreaEnd == -1))
+			switch (irpSp->MajorFunction)
 			{
-				IoCopyCurrentIrpStackLocationToNext (item->OriginalIrp);
-				IoSetCompletionRoutine (item->OriginalIrp, OnPassedIrpCompleted, item, TRUE, TRUE, TRUE);
-				IoCallDriver (queue->LowerDeviceObject, item->OriginalIrp);
+			case IRP_MJ_READ:
+				item->Write = FALSE;
+				item->OriginalOffset = irpSp->Parameters.Read.ByteOffset;
+				item->OriginalLength = irpSp->Parameters.Read.Length;
+				break;
+
+			case IRP_MJ_WRITE:
+				item->Write = TRUE;
+				item->OriginalOffset = irpSp->Parameters.Write.ByteOffset;
+				item->OriginalLength = irpSp->Parameters.Write.Length;
+				break;
+
+			default:
+				CompleteOriginalIrp (item, STATUS_INVALID_PARAMETER, 0);
+				continue;
+			}
+
+#ifdef TC_TRACE_IO_QUEUE
+			item->OriginalIrpOffset = item->OriginalOffset;
+#endif
+
+			// Handle misaligned read operations to work around a bug in Windows System Assessment Tool which does not follow FILE_FLAG_NO_BUFFERING requirements when benchmarking disk devices
+			if (queue->IsFilterDevice
+				&& !item->Write
+				&& item->OriginalLength > 0
+				&& (item->OriginalLength & (ENCRYPTION_DATA_UNIT_SIZE - 1)) == 0
+				&& (item->OriginalOffset.QuadPart & (ENCRYPTION_DATA_UNIT_SIZE - 1)) != 0)
+			{
+				byte *buffer;
+				ULONG alignedLength = item->OriginalLength + ENCRYPTION_DATA_UNIT_SIZE;
+				LARGE_INTEGER alignedOffset;
+				alignedOffset.QuadPart = item->OriginalOffset.QuadPart & ~((LONGLONG) ENCRYPTION_DATA_UNIT_SIZE - 1);
+
+				buffer = TCalloc (alignedLength);
+				if (!buffer)
+				{
+					CompleteOriginalIrp (item, STATUS_INSUFFICIENT_RESOURCES, 0);
+					continue;
+				}
+
+				item->Status = TCReadDevice (queue->LowerDeviceObject, buffer, alignedOffset, alignedLength);
+
+				if (NT_SUCCESS (item->Status))
+				{
+					UINT64_STRUCT dataUnit;
+
+					dataBuffer = (PUCHAR) MmGetSystemAddressForMdlSafe (irp->MdlAddress, HighPagePriority);
+					if (!dataBuffer)
+					{
+						TCfree (buffer);
+						CompleteOriginalIrp (item, STATUS_INSUFFICIENT_RESOURCES, 0);
+						continue;
+					}
+
+					if (queue->EncryptedAreaStart != -1 && queue->EncryptedAreaEnd != -1)
+					{
+						GetIntersection (alignedOffset.QuadPart, alignedLength, queue->EncryptedAreaStart, queue->EncryptedAreaEnd, &intersectStart, &intersectLength);
+						if (intersectLength > 0)
+						{
+							dataUnit.Value = intersectStart / ENCRYPTION_DATA_UNIT_SIZE;
+							DecryptDataUnits (buffer + (intersectStart - alignedOffset.QuadPart), &dataUnit, intersectLength / ENCRYPTION_DATA_UNIT_SIZE, queue->CryptoInfo);
+						}
+					}
+
+					memcpy (dataBuffer, buffer + (item->OriginalOffset.LowPart & (ENCRYPTION_DATA_UNIT_SIZE - 1)), item->OriginalLength);
+				}
+
+				TCfree (buffer);
+				CompleteOriginalIrp (item, item->Status, NT_SUCCESS (item->Status) ? item->OriginalLength : 0);
 				continue;
 			}
 
@@ -274,6 +605,10 @@ static VOID MainThreadProc (PVOID threadArg)
 				continue;
 			}
 
+#ifdef TC_TRACE_IO_QUEUE
+			Dump ("Q  %I64d [%I64d] %c len=%d\n", item->OriginalOffset.QuadPart, GetElapsedTime (&queue->LastPerformanceCounter), item->Write ? 'W' : 'R', item->OriginalLength);
+#endif
+
 			if (!queue->IsFilterDevice)
 			{
 				// Adjust the offset for host file or device
@@ -300,8 +635,9 @@ static VOID MainThreadProc (PVOID threadArg)
 					if (RegionsOverlap ((unsigned __int64) item->OriginalOffset.QuadPart,
 						(unsigned __int64) item->OriginalOffset.QuadPart + item->OriginalLength - 1,
 						queue->CryptoInfo->hiddenVolumeOffset,
-						(unsigned __int64) queue->VirtualDeviceLength + queue->CryptoInfo->volDataAreaOffset - (HIDDEN_VOL_HEADER_OFFSET - HEADER_SIZE) - 1))
+						(unsigned __int64) queue->CryptoInfo->hiddenVolumeOffset + queue->CryptoInfo->hiddenVolumeProtectedSize - 1))
 					{
+						Dump ("Hidden volume protection triggered: write %I64d-%I64d (protected %I64d-%I64d)\n", item->OriginalOffset.QuadPart, item->OriginalOffset.QuadPart + item->OriginalLength - 1, queue->CryptoInfo->hiddenVolumeOffset, queue->CryptoInfo->hiddenVolumeOffset + queue->CryptoInfo->hiddenVolumeProtectedSize - 1);
 						queue->CryptoInfo->bHiddenVolProtectionAction = TRUE;
 
 						// Deny this write operation to prevent the hidden volume from being overwritten
@@ -310,11 +646,25 @@ static VOID MainThreadProc (PVOID threadArg)
 					}
 				}
 			}
+			else if (item->Write
+				&& RegionsOverlap (item->OriginalOffset.QuadPart, item->OriginalOffset.QuadPart + item->OriginalLength - 1, TC_BOOT_VOLUME_HEADER_SECTOR_OFFSET, TC_BOOT_VOLUME_HEADER_SECTOR_OFFSET + TC_BOOT_ENCRYPTION_VOLUME_HEADER_SIZE - 1))
+			{
+				// Prevent inappropriately designed software from damaging important data that may be out of sync with the backup on the Rescue Disk (such as the end of the encrypted area).
+				Dump ("Preventing write to the system encryption key data area\n");
+				CompleteOriginalIrp (item, STATUS_MEDIA_WRITE_PROTECTED, 0);
+				continue;
+			}
+			else if (item->Write && IsHiddenSystemRunning()
+				&& (RegionsOverlap (item->OriginalOffset.QuadPart, item->OriginalOffset.QuadPart + item->OriginalLength - 1, TC_SECTOR_SIZE_BIOS, TC_BOOT_LOADER_AREA_SECTOR_COUNT * TC_SECTOR_SIZE_BIOS - 1)
+				 || RegionsOverlap (item->OriginalOffset.QuadPart, item->OriginalOffset.QuadPart + item->OriginalLength - 1, GetBootDriveLength(), _I64_MAX)))
+			{
+				Dump ("Preventing write to boot loader or host protected area\n");
+				CompleteOriginalIrp (item, STATUS_MEDIA_WRITE_PROTECTED, 0);
+				continue;
+			}
 
-			//Dump ("--- Queue %c %I64d  (%I64d)  %d  out=%d\n", item->Write ? 'W' : 'R', item->OriginalOffset.QuadPart, item->OriginalOffset.QuadPart / 1024 / 1024, item->OriginalLength, queue->OutstandingIoCount);
+			dataBuffer = (PUCHAR) MmGetSystemAddressForMdlSafe (irp->MdlAddress, HighPagePriority);
 
-			// Original IRP data buffer
-			dataBuffer = (PUCHAR) MmGetSystemAddressForMdlSafe (item->OriginalIrp->MdlAddress, HighPagePriority);
 			if (dataBuffer == NULL)
 			{
 				CompleteOriginalIrp (item, STATUS_INSUFFICIENT_RESOURCES, 0);
@@ -333,17 +683,10 @@ static VOID MainThreadProc (PVOID threadArg)
 				ULONG dataFragmentLength = isLastFragment ? dataRemaining : TC_ENC_IO_QUEUE_MAX_FRAGMENT_SIZE;
 				activeFragmentBuffer = (activeFragmentBuffer == queue->FragmentBufferA ? queue->FragmentBufferB : queue->FragmentBufferA);
 
+				InterlockedIncrement (&queue->IoThreadPendingRequestCount);
+
 				// Create IO request
-				request = (EncryptedIoRequest *) TCalloc (sizeof (EncryptedIoRequest));
-				if (!request)
-				{
-					while (InterlockedExchangeAdd (&item->OutstandingRequestCount, 0) > 0)
-						KeWaitForSingleObject (&queue->RequestCompletedEvent, Executive, KernelMode, FALSE, NULL);
-
-					CompleteOriginalIrp (item, STATUS_INSUFFICIENT_RESOURCES, 0);
-					break;
-				}
-
+				request = GetPoolBuffer (queue, sizeof (EncryptedIoRequest));
 				request->Item = item;
 				request->CompleteOriginalIrp = isLastFragment;
 				request->Offset = fragmentOffset;
@@ -353,11 +696,18 @@ static VOID MainThreadProc (PVOID threadArg)
 
 				if (queue->IsFilterDevice)
 				{
-					// Get intersection of data fragment with encrypted area
-					GetIntersection (fragmentOffset.QuadPart, dataFragmentLength, queue->EncryptedAreaStart, queue->EncryptedAreaEnd, &intersectStart, &intersectLength);
+					if (queue->EncryptedAreaStart == -1 || queue->EncryptedAreaEnd == -1)
+					{
+						request->EncryptedLength = 0;
+					}
+					else
+					{
+						// Get intersection of data fragment with encrypted area
+						GetIntersection (fragmentOffset.QuadPart, dataFragmentLength, queue->EncryptedAreaStart, queue->EncryptedAreaEnd, &intersectStart, &intersectLength);
 
-					request->EncryptedOffset = intersectStart - fragmentOffset.QuadPart;
-					request->EncryptedLength = intersectLength;
+						request->EncryptedOffset = intersectStart - fragmentOffset.QuadPart;
+						request->EncryptedLength = intersectLength;
+					}
 				}
 				else
 				{
@@ -381,14 +731,14 @@ static VOID MainThreadProc (PVOID threadArg)
 
 						if (queue->CryptoInfo->bPartitionInInactiveSysEncScope)
 							dataUnit.Value += queue->CryptoInfo->FirstDataUnitNo.Value;
-
+						else if (queue->RemapEncryptedArea)
+							dataUnit.Value += queue->RemappedAreaDataUnitOffset;
+								
 						EncryptDataUnits (activeFragmentBuffer + request->EncryptedOffset, &dataUnit, request->EncryptedLength / ENCRYPTION_DATA_UNIT_SIZE, queue->CryptoInfo);
 					}
 				}
 
 				// Queue IO request
-				InterlockedIncrement (&item->OutstandingRequestCount);
-
 				ExInterlockedInsertTailList (&queue->IoThreadQueue, &request->ListEntry, &queue->IoThreadQueueLock);
 				KeSetEvent (&queue->IoThreadQueueNotEmptyEvent, IO_DISK_INCREMENT, FALSE);
 
@@ -408,8 +758,6 @@ static VOID MainThreadProc (PVOID threadArg)
 
 NTSTATUS EncryptedIoQueueAddIrp (EncryptedIoQueue *queue, PIRP irp)
 {
-	EncryptedIoQueueItem *item;
-	PIO_STACK_LOCATION origIrpSp = IoGetCurrentIrpStackLocation (irp);
 	NTSTATUS status;
 
 	InterlockedIncrement (&queue->OutstandingIoCount);
@@ -420,52 +768,20 @@ NTSTATUS EncryptedIoQueueAddIrp (EncryptedIoQueue *queue, PIRP irp)
 		goto err;
 	}
 
-	if (queue->IsFilterDevice)
-	{
-		status = IoAcquireRemoveLock (&queue->RemoveLock, irp);
-		if (!NT_SUCCESS (status))
-			goto err;
-	}
-
-	item = TCalloc (sizeof (EncryptedIoQueueItem));
-	if (!item)
-	{
-		status = STATUS_INSUFFICIENT_RESOURCES;
+	status = IoAcquireRemoveLock (&queue->RemoveLock, irp);
+	if (!NT_SUCCESS (status))
 		goto err;
-	}
 
-	memset (item, 0, sizeof (EncryptedIoQueueItem));
-
-	switch (origIrpSp->MajorFunction)
+#ifdef TC_TRACE_IO_QUEUE
 	{
-	case IRP_MJ_READ:
-		item->Write = FALSE;
-		item->OriginalOffset = origIrpSp->Parameters.Read.ByteOffset;
-		item->OriginalLength = origIrpSp->Parameters.Read.Length;
-		break;
-
-	case IRP_MJ_WRITE:
-		item->Write = TRUE;
-		item->OriginalOffset = origIrpSp->Parameters.Write.ByteOffset;
-		item->OriginalLength = origIrpSp->Parameters.Write.Length;
-		break;
-
-	default:
-		TCfree (item);
-		status = STATUS_INVALID_PARAMETER;
-		goto err;
+		PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation (irp);
+		Dump ("* %I64d [%I64d] %c len=%d out=%d\n", irpSp->MajorFunction == IRP_MJ_WRITE ? irpSp->Parameters.Write.ByteOffset : irpSp->Parameters.Read.ByteOffset, GetElapsedTime (&queue->LastPerformanceCounter), irpSp->MajorFunction == IRP_MJ_WRITE ? 'W' : 'R', irpSp->MajorFunction == IRP_MJ_WRITE ? irpSp->Parameters.Write.Length : irpSp->Parameters.Read.Length, queue->OutstandingIoCount);
 	}
-
-	item->Queue = queue;
-	item->OriginalIrp = irp;
-	item->OutstandingRequestCount = 0;
-	item->Status = STATUS_SUCCESS;
+#endif
 
 	IoMarkIrpPending (irp);
 
-	//Dump ("Queue add %I64d %I64d  out=%d\n", item->OriginalOffset, item->OriginalLength, queue->OutstandingIoCount);
-
-	ExInterlockedInsertTailList (&queue->MainThreadQueue, &item->ListEntry, &queue->MainThreadQueueLock);
+	ExInterlockedInsertTailList (&queue->MainThreadQueue, &irp->Tail.Overlay.ListEntry, &queue->MainThreadQueueLock);
 	KeSetEvent (&queue->MainThreadQueueNotEmptyEvent, IO_DISK_INCREMENT, FALSE);
 	
 	return STATUS_PENDING;
@@ -496,7 +812,17 @@ NTSTATUS EncryptedIoQueueHoldWhenIdle (EncryptedIoQueue *queue, int64 timeout)
 				status = STATUS_UNSUCCESSFUL;
 
 			if (!NT_SUCCESS (status))
+			{
+				queue->SuspendPending = FALSE;
 				return status;
+			}
+
+			TCSleep (1);
+			if (InterlockedExchangeAdd (&queue->OutstandingIoCount, 0) > 0)
+			{
+				queue->SuspendPending = FALSE;
+				return STATUS_UNSUCCESSFUL;
+			}
 		}
 
 		KeClearEvent (&queue->QueueResumedEvent);
@@ -507,12 +833,11 @@ NTSTATUS EncryptedIoQueueHoldWhenIdle (EncryptedIoQueue *queue, int64 timeout)
 
 		queue->Suspended = FALSE;
 		KeSetEvent (&queue->QueueResumedEvent, IO_DISK_INCREMENT, FALSE);
-
 	}
 
-	queue->SuspendPending = FALSE;
-	//Dump ("Queue suspended  out=%d\n", queue->OutstandingIoCount);
+	queue->ReadAheadBufferValid = FALSE;
 
+	queue->SuspendPending = FALSE;
 	return STATUS_SUCCESS;
 }
 
@@ -536,19 +861,27 @@ NTSTATUS EncryptedIoQueueResumeFromHold (EncryptedIoQueue *queue)
 	queue->Suspended = FALSE;
 	KeSetEvent (&queue->QueueResumedEvent, IO_DISK_INCREMENT, FALSE);
 
-	//Dump ("Queue resumed  out=%d\n", queue->OutstandingIoCount);
-
 	return STATUS_SUCCESS;
 }
 
 
-NTSTATUS EncryptedIoQueueStart (EncryptedIoQueue *queue, PEPROCESS process)
+NTSTATUS EncryptedIoQueueStart (EncryptedIoQueue *queue)
 {
 	NTSTATUS status;
+	EncryptedIoQueueBuffer *buffer;
+	int i;
+
+	queue->StartPending = TRUE;
 	queue->ThreadExitRequested = FALSE;
 
+	queue->OutstandingIoCount = 0;
+	queue->IoThreadPendingRequestCount = 0;
+
+	queue->FirstPoolBuffer = NULL;
+	KeInitializeMutex (&queue->BufferPoolMutex, 0);
+
 	KeInitializeEvent (&queue->NoOutstandingIoEvent, SynchronizationEvent, FALSE);
-	KeInitializeEvent (&queue->RequestCompletedEvent, SynchronizationEvent, FALSE);
+	KeInitializeEvent (&queue->PoolBufferFreeEvent, SynchronizationEvent, FALSE);
 	KeInitializeEvent (&queue->QueueResumedEvent, SynchronizationEvent, FALSE);
 
 	queue->FragmentBufferA = TCalloc (TC_ENC_IO_QUEUE_MAX_FRAGMENT_SIZE);
@@ -562,6 +895,26 @@ NTSTATUS EncryptedIoQueueStart (EncryptedIoQueue *queue, PEPROCESS process)
 	KeInitializeEvent (&queue->FragmentBufferAFreeEvent, SynchronizationEvent, TRUE);
 	KeInitializeEvent (&queue->FragmentBufferBFreeEvent, SynchronizationEvent, TRUE);
 
+	queue->ReadAheadBufferValid = FALSE;
+	queue->ReadAheadBuffer = TCalloc (TC_ENC_IO_QUEUE_MAX_FRAGMENT_SIZE);
+	if (!queue->ReadAheadBuffer)
+		goto noMemory;
+
+	// Preallocate buffers
+	for (i = 0; i < TC_ENC_IO_QUEUE_PREALLOCATED_IO_REQUEST_COUNT; ++i)
+	{
+		if (i < TC_ENC_IO_QUEUE_PREALLOCATED_ITEM_COUNT && !GetPoolBuffer (queue, sizeof (EncryptedIoQueueItem)))
+			goto noMemory;
+
+		if (!GetPoolBuffer (queue, sizeof (EncryptedIoRequest)))
+			goto noMemory;
+	}
+
+	for (buffer = queue->FirstPoolBuffer; buffer != NULL; buffer = buffer->NextBuffer)
+	{
+		buffer->InUse = FALSE;
+	}
+
 	// Main thread
 	InitializeListHead (&queue->MainThreadQueue);
 	KeInitializeSpinLock (&queue->MainThreadQueueLock);
@@ -576,7 +929,7 @@ NTSTATUS EncryptedIoQueueStart (EncryptedIoQueue *queue, PEPROCESS process)
 	KeInitializeSpinLock (&queue->IoThreadQueueLock);
 	KeInitializeEvent (&queue->IoThreadQueueNotEmptyEvent, SynchronizationEvent, FALSE);
 
-	status = TCStartThreadInProcess (IoThreadProc, queue, &queue->IoThread, process);
+	status = TCStartThread (IoThreadProc, queue, &queue->IoThread);
 	if (!NT_SUCCESS (status))
 	{
 		queue->ThreadExitRequested = TRUE;
@@ -598,7 +951,13 @@ NTSTATUS EncryptedIoQueueStart (EncryptedIoQueue *queue, PEPROCESS process)
 		goto err;
 	}
 
+#ifdef TC_TRACE_IO_QUEUE
+	GetElapsedTimeInit (&queue->LastPerformanceCounter);
+#endif
+
 	queue->StopPending = FALSE;
+	queue->StartPending = FALSE;
+
 	Dump ("Queue started\n");
 	return STATUS_SUCCESS;
 
@@ -611,6 +970,9 @@ err:
 	if (queue->FragmentBufferB)
 		TCfree (queue->FragmentBufferB);
 
+	FreePoolBuffers (queue);
+
+	queue->StartPending = FALSE;
 	return status;
 }
 
@@ -635,6 +997,9 @@ NTSTATUS EncryptedIoQueueStop (EncryptedIoQueue *queue)
 
 	TCfree (queue->FragmentBufferA);
 	TCfree (queue->FragmentBufferB);
+	TCfree (queue->ReadAheadBuffer);
+
+	FreePoolBuffers (queue);
 
 	Dump ("Queue stopped  out=%d\n", queue->OutstandingIoCount);
 	return STATUS_SUCCESS;
