@@ -1,13 +1,13 @@
 /*
- Copyright (c) 2008 TrueCrypt Foundation. All rights reserved.
+ Copyright (c) 2008-2010 TrueCrypt Developers Association. All rights reserved.
 
- Governed by the TrueCrypt License 2.4 the full text of which is contained
- in the file License.txt included in TrueCrypt binary and source code
- distribution packages.
+ Governed by the TrueCrypt License 3.0 the full text of which is contained in
+ the file License.txt included in TrueCrypt binary and source code distribution
+ packages.
 */
 
 #include "System.h"
-#include "Volume/EncryptionTest.h"
+#include "Common/SecurityToken.h"
 #include "Main/Main.h"
 #include "Main/Resources.h"
 #include "Main/Application.h"
@@ -16,13 +16,13 @@
 #include "Main/Xml.h"
 #include "MainFrame.h"
 #include "AboutDialog.h"
+#include "BenchmarkDialog.h"
 #include "ChangePasswordDialog.h"
+#include "EncryptionTestDialog.h"
 #include "FavoriteVolumesDialog.h"
 #include "LegalNoticesDialog.h"
 #include "PreferencesDialog.h"
-#ifdef TC_WINDOWS
-#include "TravelerDiskWizard.h"
-#endif
+#include "SecurityTokenKeyfilesDialog.h"
 #include "VolumeCreationWizard.h"
 #include "VolumePropertiesDialog.h"
 
@@ -43,12 +43,25 @@ namespace TrueCrypt
 		InitTaskBarIcon();
 		InitEvents();
 		InitMessageFilter();
+
+		if (!GetPreferences().SecurityTokenModule.IsEmpty() && !SecurityToken::IsInitialized())
+		{
+			try
+			{
+				Gui->InitSecurityTokenLibrary();
+			}
+			catch (exception &e)
+			{
+				Gui->ShowError (e);
+			}
+		}
 	}
 
 	MainFrame::~MainFrame ()
 	{
 		Core->VolumeMountedEvent.Disconnect (this);
 		Core->VolumeDismountedEvent.Disconnect (this);
+		Gui->OpenVolumeSystemRequestEvent.Disconnect (this);
 		Gui->PreferencesUpdatedEvent.Disconnect (this);
 
 		VolumeHistory::DisconnectComboBox (VolumePathComboBox);
@@ -60,32 +73,39 @@ namespace TrueCrypt
 
 	void MainFrame::AddToFavorites (const VolumeInfoList &volumes)
 	{
-		FavoriteVolumeList newFavorites;
-
-		// Delete duplicates
-		foreach (shared_ptr <FavoriteVolume> favorite, FavoriteVolume::LoadList())
+		try
 		{
-			bool mounted = false;
+			FavoriteVolumeList newFavorites;
+
+			// Delete duplicates
+			foreach (shared_ptr <FavoriteVolume> favorite, FavoriteVolume::LoadList())
+			{
+				bool mounted = false;
+				foreach_ref (const VolumeInfo &volume, volumes)
+				{
+					if (volume.Path == favorite->Path)
+					{
+						mounted = true;
+						break;
+					}
+				}
+				if (!mounted)
+					newFavorites.push_back (favorite);
+			}
+
+			size_t newItemCount = 0;
 			foreach_ref (const VolumeInfo &volume, volumes)
 			{
-				if (volume.Path == favorite->Path)
-				{
-					mounted = true;
-					break;
-				}
+				newFavorites.push_back (shared_ptr <FavoriteVolume> (new FavoriteVolume (volume.Path, volume.MountPoint, volume.SlotNumber, volume.Protection == VolumeProtection::ReadOnly, volume.SystemEncryption)));
+				++newItemCount;
 			}
-			if (!mounted)
-				newFavorites.push_back (favorite);
-		}
 
-		size_t newItemCount = 0;
-		foreach_ref (const VolumeInfo &volume, volumes)
+			OrganizeFavorites (newFavorites, newItemCount);
+		}
+		catch (exception &e)
 		{
-			newFavorites.push_back (shared_ptr <FavoriteVolume> (new FavoriteVolume (volume.Path, volume.MountPoint, volume.SlotNumber)));
-			++newItemCount;
+			Gui->ShowError (e);
 		}
-
-		OrganizeFavorites (newFavorites, newItemCount);
 	}
 
 	bool MainFrame::CanExit () const
@@ -100,11 +120,13 @@ namespace TrueCrypt
 
 		shared_ptr <VolumePath> volumePath = GetSelectedVolumePath();
 
+#ifdef TC_WINDOWS
 		if (Core->IsVolumeMounted (*volumePath))
 		{
 			Gui->ShowInfo (LangString [mode == ChangePasswordDialog::Mode::ChangePkcs5Prf ? "MOUNTED_NO_PKCS5_PRF_CHANGE" : "MOUNTED_NOPWCHANGE"]);
 			return;
 		}
+#endif
 
 		ChangePasswordDialog dialog (this, volumePath, mode);
 		dialog.ShowModal();
@@ -127,6 +149,9 @@ namespace TrueCrypt
 					L"runas",
 					L"cmd.exe", args.c_str(), nullptr, SW_SHOW);
 #else
+#	ifdef TC_MACOSX
+				Gui->ShowInfo (_("Disk Utility will be launched after you press 'OK'.\n\nPlease select your volume in the Disk Utility window and press 'Verify Disk' or 'Repair Disk' button on the 'First Aid' page."));
+#	endif
 				Core->CheckFilesystem (selectedVolume, repair);
 				UpdateVolumeList();
 #endif
@@ -177,7 +202,6 @@ namespace TrueCrypt
 
 #ifndef TC_WINDOWS
 		SettingsMenu->Remove (HotkeysMenuItem);
-		ToolsMenu->Remove (TravelerDiskWizardMenuItem);
 #endif
 
 #ifdef TC_MACOSX
@@ -198,14 +222,14 @@ namespace TrueCrypt
 		colPermilles.push_back (75);
 #else
 		SlotListCtrl->InsertColumn (ColumnSlot, _("Slot"), wxLIST_FORMAT_LEFT, 1);
-		colPermilles.push_back (81);
+		colPermilles.push_back (82);
 #endif
 
 		SlotListCtrl->InsertColumn (ColumnPath, LangString["VOLUME"], wxLIST_FORMAT_LEFT, 1);
 #ifdef TC_WINDOWS		 
 		colPermilles.push_back (487);
 #else
-		colPermilles.push_back (430);
+		colPermilles.push_back (429);
 #endif
 
 		SlotListCtrl->InsertColumn (ColumnSize, LangString["SIZE"], wxLIST_FORMAT_RIGHT, 1);
@@ -231,22 +255,31 @@ namespace TrueCrypt
 		SlotListCtrl->AssignImageList (imageList, wxIMAGE_LIST_SMALL);
 
 		SetMinSize (wxSize (-1, -1));
-		Gui->SetListCtrlHeight (SlotListCtrl, 12);
+
+		size_t slotListRowCount = 12;
+
+#ifndef TC_WINDOWS
+		int screenHeight = wxSystemSettings::GetMetric (wxSYS_SCREEN_Y);
+		
+		if (screenHeight < 480)
+			slotListRowCount = 1;
+		else if (screenHeight <= 600)
+			slotListRowCount = slotListRowCount * screenHeight / 1000;
+#endif
+
+		Gui->SetListCtrlHeight (SlotListCtrl, slotListRowCount);
 
 #ifdef __WXGTK__
-		if (CreateVolumeButton->GetSize().GetHeight() > 26)
-		{
-			wxSize size (-1, (int) ((double) Gui->GetCharHeight (this) * 1.53));
-			CreateVolumeButton->SetMinSize (size);
-			VolumePropertiesButton->SetMinSize (size);
-			WipeCacheButton->SetMinSize (size);
-			VolumePathComboBox->SetMinSize (size);
-			SelectFileButton->SetMinSize (size);
-			SelectDeviceButton->SetMinSize (size);
-			VolumeToolsButton->SetMinSize (size);
-			size = wxSize (-1, 38);
-			VolumeButton->SetMinSize (size);
-		}
+		wxSize size (-1, (int) ((double) Gui->GetCharHeight (this) * 1.53));
+		CreateVolumeButton->SetMinSize (size);
+		VolumePropertiesButton->SetMinSize (size);
+		WipeCacheButton->SetMinSize (size);
+		VolumePathComboBox->SetMinSize (size);
+		SelectFileButton->SetMinSize (size);
+		SelectDeviceButton->SetMinSize (size);
+		VolumeToolsButton->SetMinSize (size);
+		size = wxSize (-1, 38);
+		VolumeButton->SetMinSize (size);
 #endif
 		Fit();
 		Layout();
@@ -268,6 +301,7 @@ namespace TrueCrypt
 	{
 		Core->VolumeMountedEvent.Connect (EventConnector <MainFrame> (this, &MainFrame::OnVolumeMounted));
 		Core->VolumeDismountedEvent.Connect (EventConnector <MainFrame> (this, &MainFrame::OnVolumeDismounted));
+		Gui->OpenVolumeSystemRequestEvent.Connect (EventConnector <MainFrame> (this, &MainFrame::OnOpenVolumeSystemRequestEvent));
 		Gui->PreferencesUpdatedEvent.Connect (EventConnector <MainFrame> (this, &MainFrame::OnPreferencesUpdated));
 
 		// Drag & drop
@@ -325,7 +359,7 @@ namespace TrueCrypt
 		};
 
 		mTimer.reset (dynamic_cast <wxTimer *> (new Timer (this)));
-		mTimer->Start (1000);
+		mTimer->Start (2000);
 	}
 
 #ifdef TC_WINDOWS
@@ -385,6 +419,9 @@ namespace TrueCrypt
 
 			LoadFavoriteVolumes();
 			VolumeHistory::Load();
+
+			if (VolumePathComboBox->GetValue().empty() && !VolumeHistory::Get().empty())
+				SetVolumePath (VolumeHistory::Get().front());
 		}
 		catch (exception &e)
 		{
@@ -545,11 +582,18 @@ namespace TrueCrypt
 			Gui->ShowError (e);
 		}
 	}
-	
+
 	void MainFrame::MountAllFavorites ()
 	{
-		MountOptions mountOptions (GetPreferences().DefaultMountOptions);
-		Gui->MountAllFavoriteVolumes (mountOptions);
+		try
+		{
+			MountOptions mountOptions (GetPreferences().DefaultMountOptions);
+			Gui->MountAllFavoriteVolumes (mountOptions);
+		}
+		catch (exception &e)
+		{
+			Gui->ShowError (e);
+		}
 	}
 
 	void MainFrame::MountVolume ()
@@ -586,6 +630,8 @@ namespace TrueCrypt
 
 	void MainFrame::OnActivate (wxActivateEvent& event)
 	{
+		Gui->SetActiveFrame (this);
+
 #ifdef TC_MACOSX
 		if (event.GetActive() && Gui->IsInBackgroundMode())
 			Gui->SetBackgroundMode (false);
@@ -614,35 +660,22 @@ namespace TrueCrypt
 		if (!CheckVolumePathNotEmpty ())
 			return;
 
-		shared_ptr <VolumePath> volumePath = GetSelectedVolumePath();
-
-		if (Core->IsVolumeMounted (*volumePath))
-		{
-			Gui->ShowInfo ("DISMOUNT_FIRST");
-			return;
-		}
-
-		wxString confirmMsg = LangString["CONFIRM_VOL_HEADER_BAK"];
-		confirmMsg.Replace (L"%hs", L"%s");
-
-		if (!Gui->AskYesNo (wxString::Format (confirmMsg, wstring (*volumePath).c_str()), true))
-			return;
-
-		FilePathList files = Gui->SelectFiles (this, wxEmptyString, true, false);
-		if (files.empty())
-			return;
-
 		try
 		{
-			Core->BackupVolumeHeaders (*volumePath, *files.front());
-			Gui->ShowWarning ("VOL_HEADER_BACKED_UP");
+			Gui->BackupVolumeHeaders (GetSelectedVolumePath());
 		}
-		catch (Exception &e)
+		catch (exception &e)
 		{
 			Gui->ShowError (e);
 		}
 	}
 
+	void MainFrame::OnBenchmarkMenuItemSelected (wxCommandEvent& event)
+	{
+		BenchmarkDialog dialog (this);
+		dialog.ShowModal();
+	}
+
 	void MainFrame::OnClearSlotSelectionMenuItemSelected (wxCommandEvent& event)
 	{
 		Gui->ClearListCtrlSelection (SlotListCtrl);
@@ -654,6 +687,17 @@ namespace TrueCrypt
 		if (GetPreferences().WipeCacheOnClose)
 			Core->WipePasswordCache();
 
+#ifdef TC_MACOSX
+		if (!event.CanVeto() && GetPreferences().DismountOnLogOff)
+		{
+			try
+			{
+				Gui->DismountVolumes (Core->GetMountedVolumes(), GetPreferences().ForceAutoDismount, false);
+			}
+			catch (...) { }
+		}
+#endif
+
 		if (!Gui->IsTheOnlyTopLevelWindow (this))
 		{
 			// Bring first frame to foreground
@@ -699,9 +743,32 @@ namespace TrueCrypt
 		return;
 	}
 
+	void MainFrame::OnCloseAllSecurityTokenSessionsMenuItemSelected (wxCommandEvent& event)
+	{
+		try
+		{
+			{
+				wxBusyCursor busy;
+				SecurityToken::CloseAllSessions();
+			}
+			Gui->ShowInfo ("ALL_TOKEN_SESSIONS_CLOSED");
+		}
+		catch (exception &e)
+		{
+			Gui->ShowError (e);
+		}
+	}
+
 	void MainFrame::OnCreateVolumeButtonClick (wxCommandEvent& event)
 	{
-		(new VolumeCreationWizard (nullptr))->Show();
+		try
+		{
+			(new VolumeCreationWizard (nullptr))->Show();
+		}
+		catch (exception &e)
+		{
+			Gui->ShowError (e);
+		}
 	}
 
 	void MainFrame::OnDefaultKeyfilesMenuItemSelected (wxCommandEvent& event)
@@ -710,7 +777,7 @@ namespace TrueCrypt
 		dialog.SelectPage (dialog.DefaultKeyfilesPage);
 		dialog.ShowModal();
 	}
-	
+
 	void MainFrame::OnDeviceChange (const DirectoryPath &mountPoint)
 	{
 		// Check if any host device has been removed and force dismount of volumes accordingly
@@ -743,16 +810,8 @@ namespace TrueCrypt
 
 	void MainFrame::OnEncryptionTestMenuItemSelected (wxCommandEvent& event)
 	{
-		try
-		{
-			EncryptionTest::TestAll();
-			Gui->ShowInfo ("TESTS_PASSED");
-		}
-		catch (Exception &e)
-		{
-			Gui->ShowError (e);
-			Gui->ShowError ("TESTS_FAILED");
-		}
+		EncryptionTestDialog dialog (this);
+		dialog.ShowModal();
 	}
 
 	void MainFrame::OnExitButtonClick (wxCommandEvent& event)
@@ -786,8 +845,21 @@ namespace TrueCrypt
 #ifdef TC_WINDOWS
 		switch (event.GetId())
 		{
-		case Hotkey::Id::DismountAll:
+		case Hotkey::Id::CloseAllSecurityTokenSessions:
+			try
 			{
+				SecurityToken::CloseAllSessions();
+				Gui->ShowInfo ("ALL_TOKEN_SESSIONS_CLOSED");
+			}
+			catch (exception &e) { Gui->ShowError (e); }
+			break;
+
+		case Hotkey::Id::DismountAll:
+		case Hotkey::Id::DismountAllWipeCache:
+			{
+				if (event.GetId() == Hotkey::Id::DismountAllWipeCache)
+					WipeCache();
+
 				size_t mountedCount = Core->GetMountedVolumes().size();
 				Gui->DismountAllVolumes();
 				size_t newMountedCount = Core->GetMountedVolumes().size();
@@ -841,7 +913,7 @@ namespace TrueCrypt
 
 		case Hotkey::Id::WipeCache:
 			WipeCache();
-			Gui->ShowInfo ("WIPE_CACHE");
+			Gui->ShowInfo ("PASSWORD_CACHE_WIPED");
 			break;
 
 		default:
@@ -914,11 +986,10 @@ namespace TrueCrypt
 			popup.AppendSeparator();
 			Gui->AppendToMenu (popup, _("Add to Favorites..."), this, wxCommandEventHandler (MainFrame::OnAddToFavoritesMenuItemSelected));
 
-#if defined (TC_WINDOWS) || defined (TC_LINUX)
 			popup.AppendSeparator();
 			Gui->AppendToMenu (popup, LangString["IDPM_CHECK_FILESYS"], this, wxCommandEventHandler (MainFrame::OnCheckFilesystemMenuItemSelected));
 			Gui->AppendToMenu (popup, LangString["IDPM_REPAIR_FILESYS"], this, wxCommandEventHandler (MainFrame::OnRepairFilesystemMenuItemSelected));
-#endif
+
 			popup.AppendSeparator();
 			Gui->AppendToMenu (popup, LangString["IDPM_PROPERTIES"], this, wxCommandEventHandler (MainFrame::OnVolumePropertiesButtonClick));
 
@@ -927,6 +998,14 @@ namespace TrueCrypt
 		else if (IsFreeSlotSelected())
 		{
 			Gui->AppendToMenu (popup, _("Mount Volume"), this, wxCommandEventHandler (MainFrame::OnMountVolumeMenuItemSelected));
+			
+			popup.AppendSeparator();
+
+			Gui->AppendToMenu (popup, LangString["SELECT_FILE_AND_MOUNT"], this, wxCommandEventHandler (MainFrame::OnSelectFileAndMountMenuItemSelected));
+			Gui->AppendToMenu (popup, LangString["SELECT_DEVICE_AND_MOUNT"], this, wxCommandEventHandler (MainFrame::OnSelectDeviceAndMountMenuItemSelected));
+
+			popup.AppendSeparator();
+
 			Gui->AppendToMenu (popup, _("Deselect"), this, wxCommandEventHandler (MainFrame::OnClearSlotSelectionMenuItemSelected));
 
 			PopupMenu (&popup);
@@ -953,7 +1032,20 @@ namespace TrueCrypt
 
 		UpdateControls();
 	}
-		
+	
+	void MainFrame::OnManageSecurityTokenKeyfilesMenuItemSelected (wxCommandEvent& event)
+	{
+		try
+		{
+			SecurityTokenKeyfilesDialog dialog (this, false);
+			dialog.ShowModal();
+		}
+		catch (exception &e)
+		{
+			Gui->ShowError (e);
+		}
+	}
+
 	void MainFrame::OnMountAllDevicesButtonClick (wxCommandEvent& event)
 	{
 		MountAllDevices();
@@ -972,13 +1064,24 @@ namespace TrueCrypt
 
 		if (event.IsChecked())
 		{
-			VolumeHistory::Clear();
+			try
+			{
+				VolumeHistory::Clear();
+			}
+			catch (exception &e) { Gui->ShowError (e); }
 		}
 	}
 
 	void MainFrame::OnOrganizeFavoritesMenuItemSelected (wxCommandEvent& event)
 	{
-		OrganizeFavorites (FavoriteVolume::LoadList());
+		try
+		{
+			OrganizeFavorites (FavoriteVolume::LoadList());
+		}
+		catch (exception &e)
+		{
+			Gui->ShowError (e);
+		}
 	}
 
 	void MainFrame::OnPreferencesMenuItemSelected (wxCommandEvent& event)
@@ -1005,61 +1108,35 @@ namespace TrueCrypt
 		if (!CheckVolumePathNotEmpty ())
 			return;
 
-		shared_ptr <VolumePath> volumePath = GetSelectedVolumePath();
-
-		if (Core->IsVolumeMounted (*volumePath))
-		{
-			Gui->ShowInfo ("DISMOUNT_FIRST");
-			return;
-		}
-
-		wxString confirmMsg = LangString["CONFIRM_VOL_HEADER_RESTORE"];
-		confirmMsg.Replace (L"%hs", L"%s");
-
-		if (!Gui->AskYesNo (wxString::Format (confirmMsg, wstring (*volumePath).c_str()), true, true))
-			return;
-
-		FilePathList files = Gui->SelectFiles (this, wxEmptyString, false, false);
-		if (files.empty())
-			return;
-
-		wxArrayString choices;
-		choices.Add (LangString["RESTORE_NORMAL_VOLUME_HEADER"]);
-		choices.Add (LangString["RESTORE_HIDDEN_VOLUME_HEADER"]);
-
-		wxSingleChoiceDialog choiceDialog (this, LangString["HEADER_RESTORE_TYPE"], Application::GetName(), choices);
-		choiceDialog.SetSelection (-1);
-		
-		if (choiceDialog.ShowModal() != wxID_OK)
-			return;
-
-		VolumeType::Enum volumeType;
-
-		switch (choiceDialog.GetSelection())
-		{
-		case 0:
-			volumeType = VolumeType::Normal;
-			break;
-
-		case 1:
-			volumeType = VolumeType::Hidden;
-			break;
-
-		default:
-			return;
-		}
-
 		try
 		{
-			Core->RestoreVolumeHeaders (*volumePath, volumeType, *files.front());
-			Gui->ShowInfo ("VOL_HEADER_RESTORED");
+			Gui->RestoreVolumeHeaders (GetSelectedVolumePath());
 		}
-		catch (Exception &e)
+		catch (exception &e)
 		{
 			Gui->ShowError (e);
 		}
 	}
 
+	void MainFrame::OnSecurityTokenPreferencesMenuItemSelected (wxCommandEvent& event)
+	{
+		PreferencesDialog dialog (this);
+		dialog.SelectPage (dialog.SecurityTokensPage);
+		dialog.ShowModal();
+	}
+
+	
+	void MainFrame::OnSelectDeviceAndMountMenuItemSelected (wxCommandEvent& event)
+	{
+		DevicePath path = Gui->SelectDevice (this);
+
+		if (!path.IsEmpty())
+		{
+			SetVolumePath (path);
+			OnMountVolumeMenuItemSelected (event);
+		}
+	}
+
 	void MainFrame::OnSelectDeviceButtonClick (wxCommandEvent& event)
 	{
 		DevicePath path = Gui->SelectDevice (this);
@@ -1068,6 +1145,17 @@ namespace TrueCrypt
 			SetVolumePath (path);
 	}
 
+	void MainFrame::OnSelectFileAndMountMenuItemSelected (wxCommandEvent& event)
+	{
+		FilePath path = Gui->SelectVolumeFile (this);
+
+		if (!path.IsEmpty())
+		{
+			SetVolumePath (path);
+			OnMountVolumeMenuItemSelected (event);
+		}
+	}
+
 	void MainFrame::OnSelectFileButtonClick (wxCommandEvent& event)
 	{
 		FilePath path = Gui->SelectVolumeFile (this);
@@ -1077,91 +1165,91 @@ namespace TrueCrypt
 	
 	void MainFrame::OnTimer ()
 	{
-		UpdateVolumeList();
-		UpdateWipeCacheButton();
-
-		if (GetPreferences().BackgroundTaskEnabled)
+		try
 		{
-			// Inactivity auto-dismount
-			if (GetPreferences().DismountOnInactivity)
+			UpdateVolumeList();
+			UpdateWipeCacheButton();
+
+			if (GetPreferences().BackgroundTaskEnabled)
 			{
-				VolumeInfoList inactiveVolumes;
-				wxLongLong currentTime = wxGetLocalTimeMillis().GetValue();
-
-				map <wstring, VolumeActivityMapEntry> newActivityTimeMap;
-
-				foreach (shared_ptr <VolumeInfo> volume, MountedVolumes)
+				// Inactivity auto-dismount
+				if (GetPreferences().DismountOnInactivity)
 				{
-					if (VolumeActivityMap.find (volume->Path) != VolumeActivityMap.end()
-						&& VolumeActivityMap[volume->Path].SerialInstanceNumber == volume->SerialInstanceNumber)
-					{
-						VolumeActivityMapEntry ae = VolumeActivityMap[volume->Path];
+					VolumeInfoList inactiveVolumes;
+					wxLongLong currentTime = wxGetLocalTimeMillis().GetValue();
 
-						if (volume->TotalDataRead != ae.TotalDataRead || volume->TotalDataWritten != ae.TotalDataWritten)
-						{
-							ae.LastActivityTime = currentTime;
-							ae.TotalDataRead = volume->TotalDataRead;
-							ae.TotalDataWritten = volume->TotalDataWritten;
-						}
-						else if ((currentTime - ae.LastActivityTime) > GetPreferences().MaxVolumeIdleTime * 1000LL * 60)
-						{
-							inactiveVolumes.push_back (volume);
-						}
+					map <wstring, VolumeActivityMapEntry> newActivityTimeMap;
 
-						newActivityTimeMap[volume->Path] = ae;
-					}
-					else
+					foreach (shared_ptr <VolumeInfo> volume, MountedVolumes)
 					{
-						newActivityTimeMap[volume->Path] = VolumeActivityMapEntry (*volume, currentTime);
+						if (VolumeActivityMap.find (volume->Path) != VolumeActivityMap.end()
+							&& VolumeActivityMap[volume->Path].SerialInstanceNumber == volume->SerialInstanceNumber)
+						{
+							VolumeActivityMapEntry ae = VolumeActivityMap[volume->Path];
+
+							if (volume->TotalDataRead != ae.TotalDataRead || volume->TotalDataWritten != ae.TotalDataWritten)
+							{
+								ae.LastActivityTime = currentTime;
+								ae.TotalDataRead = volume->TotalDataRead;
+								ae.TotalDataWritten = volume->TotalDataWritten;
+							}
+							else if ((currentTime - ae.LastActivityTime) > GetPreferences().MaxVolumeIdleTime * 1000LL * 60)
+							{
+								inactiveVolumes.push_back (volume);
+							}
+
+							newActivityTimeMap[volume->Path] = ae;
+						}
+						else
+						{
+							newActivityTimeMap[volume->Path] = VolumeActivityMapEntry (*volume, currentTime);
+						}
 					}
+
+					VolumeActivityMap = newActivityTimeMap;
+
+					if (!inactiveVolumes.empty())
+						Gui->AutoDismountVolumes (inactiveVolumes);
 				}
 
-				VolumeActivityMap = newActivityTimeMap;
-				
-				if (!inactiveVolumes.empty())
-					Gui->AutoDismountVolumes (inactiveVolumes);
-			}
-
-			// Screen saver auto-dismount
-			if (GetPreferences().DismountOnScreenSaver)
-			{
-#ifdef TC_WINDOWS
-				bool running;
-				if (SystemParametersInfo (SPI_GETSCREENSAVERRUNNING, 0, &running, 0) != 0)
+				// Screen saver auto-dismount
+				if (GetPreferences().DismountOnScreenSaver)
 				{
-					static bool previousState = false;
-					if (running && !previousState)
-					{
-						previousState = true;
-						Gui->OnAutoDismountAllEvent();
-					}
-					else
-					{
-						previousState = running;
-					}
-				}
-#endif
-			}
-		}
-
-		if (Gui->IsInBackgroundMode())
-		{
-			if (!GetPreferences().BackgroundTaskEnabled)
-			{
-				Close (true);
-			}
-			else if (MountedVolumes.empty() && (GetPreferences().CloseBackgroundTaskOnNoVolumes || Core->IsInTravelMode()))
-			{
-				Close (true);
-			}
-		}
-	}
-
-	void MainFrame::OnTravelerDiskWizardMenuItemSelected (wxCommandEvent& event)
-	{
 #ifdef TC_WINDOWS
-		(new TravelerDiskWizard (this))->Show();
+					bool running;
+					if (SystemParametersInfo (SPI_GETSCREENSAVERRUNNING, 0, &running, 0) != 0)
+					{
+						static bool previousState = false;
+						if (running && !previousState)
+						{
+							previousState = true;
+							Gui->OnAutoDismountAllEvent();
+						}
+						else
+						{
+							previousState = running;
+						}
+					}
 #endif
+				}
+			}
+
+			if (Gui->IsInBackgroundMode())
+			{
+				if (!GetPreferences().BackgroundTaskEnabled)
+				{
+					Close (true);
+				}
+				else if (MountedVolumes.empty() && (GetPreferences().CloseBackgroundTaskOnNoVolumes || Core->IsInPortableMode()))
+				{
+					Close (true);
+				}
+			}
+		}
+		catch (exception &e)
+		{
+			Gui->ShowError (e);
+		}
 	}
 
 	void MainFrame::OnVolumeButtonClick (wxCommandEvent& event)
@@ -1211,7 +1299,7 @@ namespace TrueCrypt
 	void MainFrame::OnWipeCacheButtonClick (wxCommandEvent& event)
 	{
 		WipeCache();
-		Gui->ShowInfo ("WIPE_CACHE");
+		Gui->ShowInfo ("PASSWORD_CACHE_WIPED");
 	}
 	
 	void MainFrame::OpenSelectedVolume () const
@@ -1254,7 +1342,7 @@ namespace TrueCrypt
 		}
 		catch (exception &e)
 		{
-			if (!Core->IsInTravelMode())
+			if (!Core->IsInPortableMode())
 				Gui->ShowError (e);
 		}
 	}
@@ -1279,7 +1367,7 @@ namespace TrueCrypt
 		{
 			wxListItem item;
 			item.SetId (itemIndex);
-			if (slotNumber == SlotListCtrl->GetItemData (item))
+			if (slotNumber == (uint32) SlotListCtrl->GetItemData (item))
 				return itemIndex;
 		}
 		return -1;
