@@ -1,9 +1,9 @@
 /*
- Copyright (c) 2008 TrueCrypt Foundation. All rights reserved.
+ Copyright (c) 2008-2010 TrueCrypt Developers Association. All rights reserved.
 
- Governed by the TrueCrypt License 2.4 the full text of which is contained
- in the file License.txt included in TrueCrypt binary and source code
- distribution packages.
+ Governed by the TrueCrypt License 3.0 the full text of which is contained in
+ the file License.txt included in TrueCrypt binary and source code distribution
+ packages.
 */
 
 #include "System.h"
@@ -13,8 +13,11 @@
 #include <wx/cmdline.h>
 #include "Platform/PlatformTest.h"
 #ifdef TC_UNIX
+#include <errno.h>
 #include "Platform/Unix/Process.h"
 #endif
+#include "Platform/SystemInfo.h"
+#include "Common/SecurityToken.h"
 #include "Volume/EncryptionTest.h"
 #include "Application.h"
 #include "FavoriteVolume.h"
@@ -30,6 +33,24 @@ namespace TrueCrypt
 	{
 		Core->WarningEvent.Disconnect (this);
 		Core->VolumeMountedEvent.Disconnect (this);
+
+		try
+		{
+			if (SecurityToken::IsInitialized())
+				SecurityToken::CloseLibrary();
+		}
+		catch (...) { }
+	}
+
+	void UserInterface::CheckRequirementsForMountingVolume () const
+	{
+#ifdef TC_LINUX
+		if (!Preferences.NonInteractive)
+		{
+			if (!SystemInfo::IsVersionAtLeast (2, 6, 24))
+				ShowWarning (_("Your system uses an old version of the Linux kernel.\n\nDue to a bug in the Linux kernel, your system may stop responding when writing data to a TrueCrypt volume. This problem can be solved by upgrading the kernel to version 2.6.24 or later."));
+		}
+#endif // TC_LINUX
 	}
 
 	void UserInterface::CloseExplorerWindows (shared_ptr <VolumeInfo> mountedVolume) const
@@ -124,6 +145,8 @@ namespace TrueCrypt
 	{
 		BusyScope busy (this);
 
+		volumes.sort (VolumeInfo::FirstVolumeMountedAfterSecond);
+
 		wxString message;
 		bool twoPassMode = volumes.size() > 1;
 		bool volumesInUse = false;
@@ -144,7 +167,7 @@ namespace TrueCrypt
 				try
 				{
 					BusyScope busy (this);
-					Core->DismountVolume (volume, ignoreOpenFiles);
+					volume = Core->DismountVolume (volume, ignoreOpenFiles);
 				}
 				catch (MountedVolumeInUse&)
 				{
@@ -162,7 +185,7 @@ namespace TrueCrypt
 						if (AskYesNo (StringFormatter (LangString["UNMOUNT_LOCK_FAILED"], wstring (volume->Path)), true, true))
 						{
 							BusyScope busy (this);
-							Core->DismountVolume (volume, true);
+							volume = Core->DismountVolume (volume, true);
 						}
 						else
 							throw UserAbort (SRC_POS);
@@ -231,7 +254,7 @@ namespace TrueCrypt
 
 			wxString protection;
 			if (volume.Type == VolumeType::Hidden)
-				protection = LangString["N_A_UISTR"];
+				protection = LangString["NOT_APPLICABLE_OR_NOT_AVAILABLE"];
 			else if (volume.HiddenVolumeProtectionTriggered)
 				protection = LangString["HID_VOL_DAMAGE_PREVENTED"];
 			else
@@ -241,6 +264,9 @@ namespace TrueCrypt
 			prop << LangString["ENCRYPTION_ALGORITHM"] << L": " << volume.EncryptionAlgorithmName << L'\n';
 			prop << LangString["KEY_SIZE"] << L": " << StringFormatter (L"{0} {1}", volume.EncryptionAlgorithmKeySize * 8, LangString ["BITS"]) << L'\n';
 
+			if (volume.EncryptionModeName == L"XTS")
+				prop << LangString["SECONDARY_KEY_SIZE_XTS"] << L": " << StringFormatter (L"{0} {1}", volume.EncryptionAlgorithmKeySize * 8, LangString ["BITS"]) << L'\n';;
+
 			wstringstream blockSize;
 			blockSize << volume.EncryptionAlgorithmBlockSize * 8;
 			if (volume.EncryptionAlgorithmBlockSize != volume.EncryptionAlgorithmMinBlockSize)
@@ -249,13 +275,19 @@ namespace TrueCrypt
 			prop << LangString["BLOCK_SIZE"] << L": " << blockSize.str() + L" " + LangString ["BITS"] << L'\n';
 			prop << LangString["MODE_OF_OPERATION"] << L": " << volume.EncryptionModeName << L'\n';
 			prop << LangString["PKCS5_PRF"] << L": " << volume.Pkcs5PrfName << L'\n';
-			prop << LangString["PKCS5_ITERATIONS"] << L": " << StringConverter::FromNumber (volume.Pkcs5IterationCount) << L'\n';
-
-			prop << LangString["VOLUME_CREATE_DATE"] << L": " << VolumeTimeToString (volume.VolumeCreationTime) << L'\n';
-			prop << LangString["VOLUME_HEADER_DATE"] << L": " << VolumeTimeToString (volume.HeaderCreationTime) << L'\n';
+	
+			prop << LangString["VOLUME_FORMAT_VERSION"] << L": " << (volume.MinRequiredProgramVersion < 0x600 ? 1 : 2) << L'\n';
+			prop << LangString["BACKUP_HEADER"] << L": " << LangString[volume.MinRequiredProgramVersion >= 0x600 ? "UISTR_YES" : "UISTR_NO"] << L'\n';
 
+#ifdef TC_LINUX
+			if (string (volume.VirtualDevice).find ("/dev/mapper/truecrypt") != 0)
+			{
+#endif
 			prop << LangString["TOTAL_DATA_READ"] << L": " << SizeToString (volume.TotalDataRead) << L'\n';
 			prop << LangString["TOTAL_DATA_WRITTEN"] << L": " << SizeToString (volume.TotalDataWritten) << L'\n';
+#ifdef TC_LINUX
+			}
+#endif
 		
 			prop << L'\n';
 		}
@@ -301,6 +333,12 @@ namespace TrueCrypt
 
 					message << e->GetSubject();
 				}
+
+#ifdef TC_UNIX
+				if (sysEx && sysEx->GetErrorCode() == EIO)
+					message << L"\n\n" << LangString["ERR_HARDWARE_ERROR"];
+#endif
+
 #ifdef DEBUG
 				if (sysEx && sysEx->what())
 					message << L"\n\n" << StringConverter::ToWide (sysEx->what());
@@ -352,20 +390,38 @@ namespace TrueCrypt
 		}
 
 		// PasswordIncorrect 
-		if (dynamic_cast <const PasswordIncorrect *> (&ex))
+		if (dynamic_cast <const PasswordException *> (&ex))
 		{
 			wxString message = ExceptionTypeToString (typeid (ex));
 
 #ifndef TC_NO_GUI
-#ifdef __WXGTK__
-			if (Application::GetUserInterfaceType() != UserInterfaceType::Text)
-#endif
-			if (wxGetKeyState (WXK_CAPITAL))
+			if (Application::GetUserInterfaceType() == UserInterfaceType::Graphic && wxGetKeyState (WXK_CAPITAL))
 				message += wxString (L"\n\n") + LangString["CAPSLOCK_ON"];
 #endif
 			return message;
 		}
 
+		// PKCS#11 Exception
+		if (dynamic_cast <const Pkcs11Exception *> (&ex))
+		{
+			string errorString = string (dynamic_cast <const Pkcs11Exception &> (ex));
+			
+			if (LangString.Exists (errorString))
+				return LangString[errorString];
+
+			if (errorString.find ("CKR_") == 0)
+			{
+				errorString = errorString.substr (4);
+				for (size_t i = 0; i < errorString.size(); ++i)
+				{
+					if (errorString[i] == '_')
+						errorString[i] = ' ';
+				}
+			}
+
+			return LangString["SECURITY_TOKEN_ERROR"] + L":\n\n" + StringConverter::ToWide (errorString);
+		}
+
 		// Other library exceptions
 		return ExceptionTypeToString (typeid (ex));
 	}
@@ -374,30 +430,54 @@ namespace TrueCrypt
 	{
 #define EX2MSG(exception, message) do { if (ex == typeid (exception)) return (message); } while (false)
 		EX2MSG (DriveLetterUnavailable,				LangString["DRIVE_LETTER_UNAVAILABLE"]);
+		EX2MSG (EncryptedSystemRequired,			_("This operation must be performed only when the system hosted on the volume is running."));
 		EX2MSG (ExternalException,					LangString["EXCEPTION_OCCURRED"]);
 		EX2MSG (InsufficientData,					_("Not enough data available."));
+		EX2MSG (InvalidSecurityTokenKeyfilePath,	LangString["INVALID_TOKEN_KEYFILE_PATH"]);
 		EX2MSG (HigherVersionRequired,				LangString["NEW_VERSION_REQUIRED"]);
+		EX2MSG (KernelCryptoServiceTestFailed,		_("Kernel cryptographic service test failed. The cryptographic service of your kernel most likely does not support volumes larger than 2 TB.\n\nPossible solutions:\n- Upgrade the Linux kernel to version 2.6.33 or later.\n- Disable use of the kernel cryptographic services (Settings > Preferences > System Integration) or use 'nokernelcrypto' mount option on the command line."));
+		EX2MSG (KeyfilePathEmpty,					LangString["ERR_KEYFILE_PATH_EMPTY"]);
+		EX2MSG (LoopDeviceSetupFailed,				_("Failed to set up a loop device."));
 		EX2MSG (MissingArgument,					_("A required argument is missing."));
 		EX2MSG (MissingVolumeData,					_("Volume data missing."));
 		EX2MSG (MountPointRequired,					_("Mount point required."));
 		EX2MSG (MountPointUnavailable,				_("Mount point is already in use."));
 		EX2MSG (NoDriveLetterAvailable,				LangString["NO_FREE_DRIVES"]);
-		EX2MSG (NoLoopbackDeviceAvailable,			_("No loopback device available."));
 		EX2MSG (PasswordEmpty,						_("No password or keyfile specified."));
 		EX2MSG (PasswordIncorrect,					LangString["PASSWORD_WRONG"]);
 		EX2MSG (PasswordKeyfilesIncorrect,			LangString["PASSWORD_OR_KEYFILE_WRONG"]);
+		EX2MSG (PasswordOrKeyboardLayoutIncorrect,	LangString["PASSWORD_OR_KEYFILE_WRONG"] + _("\n\nNote that pre-boot authentication passwords need to be typed in the pre-boot environment where non-US keyboard layouts are not available. Therefore, pre-boot authentication passwords must always be typed using the standard US keyboard layout (otherwise, the password will be typed incorrectly in most cases). However, note that you do NOT need a real US keyboard; you just need to change the keyboard layout in your operating system."));
+		EX2MSG (PasswordOrMountOptionsIncorrect,	LangString["PASSWORD_OR_KEYFILE_OR_MODE_WRONG"] + _("\n\nNote: If you are attempting to mount a partition located on an encrypted system drive without pre-boot authentication or to mount the encrypted system partition of an operating system that is not running, you can do so by selecting 'Options >' > 'Mount partition using system encryption'."));
 		EX2MSG (PasswordTooLong,					StringFormatter (_("Password is longer than {0} characters."), (int) VolumePassword::MaxSize));
-		EX2MSG (ProtectionPasswordIncorrect,		_("Incorrect keyfile(s) and/or password to the protected hidden volume or the hidden volume does not exist."));
-		EX2MSG (ProtectionPasswordKeyfilesIncorrect,	_("Incorrect password to the protected hidden volume or the hidden volume does not exist."));
+		EX2MSG (PartitionDeviceRequired,			_("Partition device required."));
+		EX2MSG (ProtectionPasswordIncorrect,		_("Incorrect password to the protected hidden volume or the hidden volume does not exist."));
+		EX2MSG (ProtectionPasswordKeyfilesIncorrect,_("Incorrect keyfile(s) and/or password to the protected hidden volume or the hidden volume does not exist."));
 		EX2MSG (RootDeviceUnavailable,				LangString["NODRIVER"]);
+		EX2MSG (SecurityTokenKeyfileAlreadyExists,	LangString["TOKEN_KEYFILE_ALREADY_EXISTS"]);
+		EX2MSG (SecurityTokenKeyfileNotFound,		LangString["TOKEN_KEYFILE_NOT_FOUND"]);
+		EX2MSG (SecurityTokenLibraryNotInitialized,	LangString["PKCS11_MODULE_INIT_FAILED"]);
 		EX2MSG (StringConversionFailed,				_("Invalid characters encountered."));
 		EX2MSG (StringFormatterException,			_("Error while parsing formatted string."));
+		EX2MSG (TemporaryDirectoryFailure,			_("Failed to create a file or directory in a temporary directory.\n\nPlease make sure that the temporary directory exists, its security permissions allow you to access it, and there is sufficient disk space."));
 		EX2MSG (UnportablePassword,					LangString["UNSUPPORTED_CHARS_IN_PWD"]);
-		EX2MSG (UnsupportedSectorSize,					LangString["LARGE_SECTOR_UNSUPPORTED"]);
+
+#if defined (TC_LINUX)
+		EX2MSG (UnsupportedSectorSize,				LangString["SECTOR_SIZE_UNSUPPORTED"]);
+		EX2MSG (UnsupportedSectorSizeHiddenVolumeProtection, _("Error: The drive uses a sector size other than 512 bytes.\n\nDue to limitations of components available on your platform, outer volumes hosted on the drive cannot be mounted using hidden volume protection.\n\nPossible solutions:\n- Use a drive with 512-byte sectors.\n- Create a file-hosted volume (container) on the drive.\n- Backup the contents of the hidden volume and then update the outer volume."));
+		EX2MSG (UnsupportedSectorSizeNoKernelCrypto, _("Error: The drive uses a sector size other than 512 bytes.\n\nDue to limitations of components available on your platform, partition/device-hosted volumes on the drive can only be mounted using kernel cryptographic services.\n\nPossible solutions:\n- Enable use of the kernel cryptographic services (Preferences > System Integration).\n- Use a drive with 512-byte sectors.\n- Create a file-hosted volume (container) on the drive."));
+#else
+		EX2MSG (UnsupportedSectorSize,				_("Error: The drive uses a sector size other than 512 bytes.\n\nDue to limitations of components available on your platform, partition/device-hosted volumes cannot be created/used on the drive.\n\nPossible solutions:\n- Create a file-hosted volume (container) on the drive.\n- Use a drive with 512-byte sectors.\n- Use TrueCrypt on another platform."));
+#endif
+
 		EX2MSG (VolumeAlreadyMounted,				LangString["VOL_ALREADY_MOUNTED"]);
+		EX2MSG (VolumeEncryptionNotCompleted,		LangString["ERR_ENCRYPTION_NOT_COMPLETED"]);
 		EX2MSG (VolumeHostInUse,					_("The host file/device is already in use."));
 		EX2MSG (VolumeSlotUnavailable,				_("Volume slot unavailable."));
 
+#ifdef TC_MACOSX
+		EX2MSG (HigherFuseVersionRequired,			_("TrueCrypt requires MacFUSE 1.3 or later."));
+#endif
+
 #undef EX2MSG
 		return L"";
 	}
@@ -425,7 +505,7 @@ namespace TrueCrypt
 		{
 			struct AdminPasswordRequestHandler : public GetStringFunctor
 			{
-				virtual string operator() ()
+				virtual void operator() (string &str)
 				{
 					throw ElevationFailed (SRC_POS, "sudo", 1, "");
 				}
@@ -436,6 +516,21 @@ namespace TrueCrypt
 
 		Core->WarningEvent.Connect (EventConnector <UserInterface> (this, &UserInterface::OnWarning));
 		Core->VolumeMountedEvent.Connect (EventConnector <UserInterface> (this, &UserInterface::OnVolumeMounted));
+
+		if (!CmdLine->Preferences.SecurityTokenModule.IsEmpty() && !SecurityToken::IsInitialized())
+		{
+			try
+			{
+				InitSecurityTokenLibrary();
+			}
+			catch (exception &e)
+			{
+				if (Preferences.NonInteractive)
+					throw;
+
+				ShowError (e);
+			}
+		}
 	}
 	
 	void UserInterface::ListMountedVolumes (const VolumeInfoList &volumes) const
@@ -559,10 +654,13 @@ namespace TrueCrypt
 				ShowInfo (LangString[newMountedVolumes.size() > 1 ? "HIDVOL_PROT_WARN_AFTER_MOUNT_PLURAL" : "HIDVOL_PROT_WARN_AFTER_MOUNT"]);
 		}
 
+		if (!newMountedVolumes.empty() && GetPreferences().CloseSecurityTokenSessionsAfterMount)
+			SecurityToken::CloseAllSessions();
+
 		return newMountedVolumes;
 	}
 
-	VolumeInfoList UserInterface::MountAllFavoriteVolumes (MountOptions &options) const
+	VolumeInfoList UserInterface::MountAllFavoriteVolumes (MountOptions &options)
 	{
 		BusyScope busy (this);
 		
@@ -593,7 +691,15 @@ namespace TrueCrypt
 				}
 				catch (...)
 				{
+					UserPreferences prefs = GetPreferences();
+					if (prefs.CloseSecurityTokenSessionsAfterMount)
+						Preferences.CloseSecurityTokenSessionsAfterMount = false;
+
 					shared_ptr <VolumeInfo> volume = MountVolume (options);
+
+					if (prefs.CloseSecurityTokenSessionsAfterMount)
+						Preferences.CloseSecurityTokenSessionsAfterMount = true;
+
 					if (!volume)
 						break;
 					newMountedVolumes.push_back (volume);
@@ -601,6 +707,9 @@ namespace TrueCrypt
 			}
 		}
 
+		if (!newMountedVolumes.empty() && GetPreferences().CloseSecurityTokenSessionsAfterMount)
+			SecurityToken::CloseAllSessions();
+
 		return newMountedVolumes;
 	}
 
@@ -640,6 +749,9 @@ namespace TrueCrypt
 		if (options.Protection == VolumeProtection::HiddenVolumeReadOnly)
 			ShowInfo ("HIDVOL_PROT_WARN_AFTER_MOUNT");
 
+		if (GetPreferences().CloseSecurityTokenSessionsAfterMount)
+			SecurityToken::CloseAllSessions();
+
 		return volume;
 	}
 
@@ -696,11 +808,17 @@ namespace TrueCrypt
 #elif defined (TC_MACOSX)
 
 		args.push_back (string (path));
-		Process::Execute ("open", args);
+		try
+		{
+			Process::Execute ("open", args);
+		}
+		catch (exception &e) { ShowError (e); }
 
 #else
 		// MIME handler for directory seems to be unavailable through wxWidgets
-		if (GetTraits()->GetDesktopEnvironment() == L"GNOME")
+		wxString desktop = GetTraits()->GetDesktopEnvironment();
+
+		if (desktop == L"GNOME" || desktop.empty())
 		{
 			args.push_back ("--no-default-window");
 			args.push_back ("--no-desktop");
@@ -710,16 +828,28 @@ namespace TrueCrypt
 				Process::Execute ("nautilus", args, 2000);
 			}
 			catch (TimeOut&) { }
+			catch (exception &e) { ShowError (e); }
 		}
-		else if (GetTraits()->GetDesktopEnvironment() == L"KDE")
+		else if (desktop == L"KDE")
 		{
-			args.push_back ("openURL");
-			args.push_back (string (path));
 			try
 			{
-				Process::Execute ("kfmclient", args, 2000);
+				args.push_back (string (path));
+				Process::Execute ("dolphin", args, 2000);
 			}
 			catch (TimeOut&) { }
+			catch (exception&)
+			{
+				args.clear();
+				args.push_back ("openURL");
+				args.push_back (string (path));
+				try
+				{
+					Process::Execute ("kfmclient", args, 2000);
+				}
+				catch (TimeOut&) { }
+				catch (exception &e) { ShowError (e); }
+			}
 		}
 #endif
 	}
@@ -820,10 +950,18 @@ namespace TrueCrypt
 			}
 			return true;
 
+		case CommandId::BackupHeaders:
+			BackupVolumeHeaders (cmdLine.ArgVolumePath);
+			return true;
+
 		case CommandId::ChangePassword:
 			ChangePassword (cmdLine.ArgVolumePath, cmdLine.ArgPassword, cmdLine.ArgKeyfiles, cmdLine.ArgNewPassword, cmdLine.ArgNewKeyfiles, cmdLine.ArgHash);
 			return true;
 
+		case CommandId::CreateKeyfile:
+			CreateKeyfile (cmdLine.ArgFilePath);
+			return true;
+
 		case CommandId::CreateVolume:
 			{
 				make_shared_auto (VolumeCreationOptions, options);
@@ -839,16 +977,20 @@ namespace TrueCrypt
 				options->Keyfiles = cmdLine.ArgKeyfiles;
 				options->Password = cmdLine.ArgPassword;
 				options->Quick = cmdLine.ArgQuick;
-				options->Size = (cmdLine.ArgSize + SECTOR_SIZE - 1) & ~(SECTOR_SIZE - 1);
+				options->Size = cmdLine.ArgSize;
 				options->Type = cmdLine.ArgVolumeType;
 
 				if (cmdLine.ArgVolumePath)
 					options->Path = VolumePath (*cmdLine.ArgVolumePath);
 
-				CreateVolume (options, cmdLine.ArgRandomSourcePath);
+				CreateVolume (options);
 				return true;
 			}
 
+		case CommandId::DeleteSecurityTokenKeyfiles:
+			DeleteSecurityTokenKeyfiles();
+			return true;
+
 		case CommandId::DismountVolumes:
 			DismountVolumes (cmdLine.ArgVolumes, cmdLine.ArgForce, !Preferences.NonInteractive);
 			return true;
@@ -875,31 +1017,69 @@ namespace TrueCrypt
 					"--auto-mount=devices|favorites\n"
 					" Auto mount device-hosted or favorite volumes.\n"
 					"\n"
-					"-c, --create=[VOLUME_PATH]\n"
-					" Create a new volume. Most options are requested from user if not specified\n"
+					"--backup-headers[=VOLUME_PATH]\n"
+					" Backup volume headers to a file. All required options are requested from the\n"
+					" user.\n"
+					"\n"
+					"-c, --create[=VOLUME_PATH]\n"
+					" Create a new volume. Most options are requested from the user if not specified\n"
 					" on command line. See also options --encryption, -k, --filesystem, --hash, -p,\n"
 					" --random-source, --quick, --size, --volume-type. Note that passing some of the\n"
-					" options may affect security (see option -p for more information).\n"
+					" options may affect security of the volume (see option -p for more information).\n"
 					"\n"
-					"-C, --change=[VOLUME_PATH]\n"
-					" Change a password and/or keyfile(s) of a volume. Volume path and passwords are\n"
-					" requested from user if not specified on command line. PKCS-5 PRF HMAC hash\n"
+					" Inexperienced users should use the graphical user interface to create a hidden\n"
+					" volume. When using the text user interface, the following procedure must be\n"
+					" followed to create a hidden volume:\n"
+					"  1) Create an outer volume with no filesystem.\n"
+					"  2) Create a hidden volume within the outer volume.\n"
+					"  3) Mount the outer volume using hidden volume protection.\n"
+					"  4) Create a filesystem on the virtual device of the outer volume.\n"
+					"  5) Mount the new filesystem and fill it with data.\n"
+					"  6) Dismount the outer volume.\n"
+					"  If at any step the hidden volume protection is triggered, start again from 1).\n"
+					"\n"
+					"--create-keyfile[=FILE_PATH]\n"
+					" Create a new keyfile containing pseudo-random data.\n"
+					"\n"
+					"-C, --change[=VOLUME_PATH]\n"
+					" Change a password and/or keyfile(s) of a volume. Most options are requested\n"
+					" from the user if not specified on command line. PKCS-5 PRF HMAC hash\n"
 					" algorithm can be changed with option --hash. See also options -k,\n"
-					" --new-keyfiles, --new-password, -p, --random-source, -v.\n"
+					" --new-keyfiles, --new-password, -p, --random-source.\n"
 					"\n"
-					"-d, --dismount=[MOUNTED_VOLUME]\n"
+					"-d, --dismount[=MOUNTED_VOLUME]\n"
 					" Dismount a mounted volume. If MOUNTED_VOLUME is not specified, all\n"
-					" volumes are dismounted. See below for a description of MOUNTED_VOLUME.\n"
+					" volumes are dismounted. See below for description of MOUNTED_VOLUME.\n"
 					"\n"
-					"-l, --list=[MOUNTED_VOLUME]\n"
+					"--delete-token-keyfiles\n"
+					" Delete keyfiles from security tokens. See also command --list-token-keyfiles.\n"
+					"\n"
+					"--export-token-keyfile\n"
+					" Export a keyfile from a security token. See also command --list-token-keyfiles.\n"
+					"\n"
+					"--import-token-keyfiles\n"
+					" Import keyfiles to a security token. See also option --token-lib.\n"
+					"\n"
+					"-l, --list[=MOUNTED_VOLUME]\n"
 					" Display a list of mounted volumes. If MOUNTED_VOLUME is not specified, all\n"
 					" volumes are listed. By default, the list contains only volume path, virtual\n"
 					" device, and mount point. A more detailed list can be enabled by verbose\n"
-					" output option (-v). See below for a description of MOUNTED_VOLUME.\n"
+					" output option (-v). See below for description of MOUNTED_VOLUME.\n"
 					"\n"
-					"--mount=[VOLUME_PATH]\n"
-					" Mount a volume interactively. Options which may affect security are\n"
-					" requested from the user. See option -p for more information.\n"
+					"--list-token-keyfiles\n"
+					" Display a list of all available security token keyfiles. See also command\n"
+					" --import-token-keyfiles.\n"
+					"\n"
+					"--mount[=VOLUME_PATH]\n"
+					" Mount a volume. Volume path and other options are requested from the user\n"
+					" if not specified on command line.\n"
+					"\n"
+					"--restore-headers[=VOLUME_PATH]\n"
+					" Restore volume headers from the embedded or an external backup. All required\n"
+					" options are requested from the user.\n"
+					"\n"
+					"--save-preferences\n"
+					" Save user preferences.\n"
 					"\n"
 					"--test\n"
 					" Test internal algorithms used in the process of encryption and decryption.\n"
@@ -907,8 +1087,8 @@ namespace TrueCrypt
 					"--version\n"
 					" Display program version.\n"
 					"\n"
-					"--volume-properties=[MOUNTED_VOLUME]\n"
-					" Display properties of a mounted volume. See below for a description of\n"
+					"--volume-properties[=MOUNTED_VOLUME]\n"
+					" Display properties of a mounted volume. See below for description of\n"
 					" MOUNTED_VOLUME.\n"
 					"\n"
 					"MOUNTED_VOLUME:\n"
@@ -920,6 +1100,9 @@ namespace TrueCrypt
 					"\n"
 					"Options:\n"
 					"\n"
+					"--display-password\n"
+					" Display password characters while typing.\n"
+					"\n"
 					"--encryption=ENCRYPTION_ALGORITHM\n"
 					" Use specified encryption algorithm when creating a new volume.\n"
 					"\n"
@@ -935,23 +1118,45 @@ namespace TrueCrypt
 					"\n"
 					"--fs-options=OPTIONS\n"
 					" Filesystem mount options. The OPTIONS argument is passed to mount(8)\n"
-					" command with option -o. This option is not available on some platforms.\n"
+					" command with option -o when a filesystem on a TrueCrypt volume is mounted.\n"
+					" This option is not available on some platforms.\n"
 					"\n"
 					"--hash=HASH\n"
 					" Use specified hash algorithm when creating a new volume or changing password\n"
-					" and/or keyfiles.\n"
+					" and/or keyfiles. This option also specifies the mixing PRF of the random\n"
+					" number generator.\n"
 					"\n"
-					"-k, --keyfiles=KEYFILE1,KEYFILE2,KEYFILE3,..\n"
-					" Use specified keyfiles when mounting a volume (or when changing password\n"
-					" and/or keyfiles). When a directory is specified, all files inside it will be\n"
+					"-k, --keyfiles=KEYFILE1[,KEYFILE2,KEYFILE3,...]\n"
+					" Use specified keyfiles when mounting a volume or when changing password\n"
+					" and/or keyfiles. When a directory is specified, all files inside it will be\n"
 					" used (non-recursively). Multiple keyfiles must be separated by comma.\n"
-					" An empty keyfile (-k \"\") disables interactive requests for keyfiles. See also\n"
-					" options --new-keyfiles, --protection-keyfiles.\n"
+					" Use double comma (,,) to specify a comma contained in keyfile's name.\n"
+					" Keyfile stored on a security token must be specified as\n"
+					" token://slot/SLOT_NUMBER/file/FILENAME. An empty keyfile (-k \"\") disables\n"
+					" interactive requests for keyfiles. See also options --import-token-keyfiles,\n"
+					" --list-token-keyfiles, --new-keyfiles, --protection-keyfiles.\n"
 					"\n"
-					"-m, --mount-options=readonly|ro|timestamp|ts\n"
-					" Specify comma-separated mount options for a TrueCrypt volume:\n"
-					"  readonly|ro: Mount read-only.\n"
-					"  timestamp|ts: Update (do not preserve) host-file timestamps.\n"
+					"--load-preferences\n"
+					" Load user preferences.\n"
+					"\n"
+					"-m, --mount-options=OPTION1[,OPTION2,OPTION3,...]\n"
+					" Specifies comma-separated mount options for a TrueCrypt volume:\n"
+					"  headerbak: Use backup headers when mounting a volume.\n"
+					"  nokernelcrypto: Do not use kernel cryptographic services.\n"
+					"  readonly|ro: Mount volume as read-only.\n"
+					"  system: Mount partition using system encryption.\n"
+					"  timestamp|ts: Do not restore host-file modification timestamp when a volume\n"
+					"   is dismounted (note that the operating system under certain circumstances\n"
+					"   does not alter host-file timestamps, which may be mistakenly interpreted\n"
+					"   to mean that this option does not work).\n"
+					" See also option --fs-options.\n"
+					"\n"
+					"--new-keyfiles=KEYFILE1[,KEYFILE2,KEYFILE3,...]\n"
+					" Add specified keyfiles to a volume. This option can only be used with command\n"
+					" -C.\n"
+					"\n"
+					"--new-password=PASSWORD\n"
+					" Specifies a new password. This option can only be used with command -C.\n"
 					"\n"
 					"-p, --password=PASSWORD\n"
 					" Use specified password to mount/open a volume. An empty password can also be\n"
@@ -963,14 +1168,14 @@ namespace TrueCrypt
 					" Write-protect a hidden volume when mounting an outer volume. Before mounting\n"
 					" the outer volume, the user will be prompted for a password to open the hidden\n"
 					" volume. The size and position of the hidden volume is then determined and the\n"
-					" outer volume is mapped with all sectors belonging to the hidden volume\n"
+					" outer volume is mounted with all sectors belonging to the hidden volume\n"
 					" protected against write operations. When a write to the protected area is\n"
 					" prevented, the whole volume is switched to read-only mode. Verbose list\n"
 					" (-v -l) can be used to query the state of the hidden volume protection.\n"
 					" Warning message is displayed when a volume switched to read-only is being\n"
-					" dismounted. See also option --mount.\n"
+					" dismounted.\n"
 					"\n"
-					"--protection-keyfiles=KEYFILE1,KEYFILE2,KEYFILE3,..\n"
+					"--protection-keyfiles=KEYFILE1[,KEYFILE2,KEYFILE3,...]\n"
 					" Use specified keyfiles to open a hidden volume to be protected. This option\n"
 					" may be used only when mounting an outer volume with hidden volume protected.\n"
 					" See also options -k and --protect-hidden.\n"
@@ -981,11 +1186,12 @@ namespace TrueCrypt
 					" See also options -p and --protect-hidden.\n"
 					"\n"
 					"--quick\n"
-					" Use quick format when creating a new volume. This option can be used only\n"
-					" when creating a device-hosted volume.\n"
+					" Do not encrypt free space when creating a device-hosted volume. This option\n"
+					" must not be used when creating an outer volume.\n"
 					"\n"
 					"--random-source=FILE\n"
-					" Use FILE as a source of random data (e.g., when creating a volume).\n"
+					" Use FILE as a source of random data (e.g., when creating a volume) instead\n"
+					" of requiring the user to type random characters.\n"
 					"\n"
 					"--slot=SLOT\n"
 					" Use specified slot number when mounting, dismounting, or listing a volume.\n"
@@ -995,23 +1201,33 @@ namespace TrueCrypt
 					"\n"
 					"-t, --text\n"
 					" Use text user interface. Graphical user interface is used by default if\n"
-					" available.\n"
+					" available. This option must be specified as the first argument.\n"
+					"\n"
+					"--token-lib=LIB_PATH\n"
+					" Use specified PKCS #11 security token library.\n"
 					"\n"
 					"--volume-type=TYPE\n"
 					" Use specified volume type when creating a new volume. TYPE can be 'normal'\n"
-					" or 'hidden'.\n"
+					" or 'hidden'. See option -c for more information on creating hidden volumes.\n"
 					"\n"
 					"-v, --verbose\n"
 					" Enable verbose output.\n"
 					"\n"
+					"\n"
+					"IMPORTANT:\n"
+					"\n"
+					"If you want to use TrueCrypt, you must follow the security requirements and\n"
+					"security precautions listed in chapter 'Security Requirements and Precautions'\n"
+					"in the TrueCrypt documentation (file 'TrueCrypt User Guide.pdf').\n"
+					"\n"
 					"\nExamples:\n\n"
-					"Create a new volume using text interface:\n"
+					"Create a new volume:\n"
 					"truecrypt -t -c\n"
 					"\n"
 					"Mount a volume:\n"
 					"truecrypt volume.tc /media/truecrypt1\n"
 					"\n"
-					"Mount a volume read-only, using keyfiles:\n"
+					"Mount a volume as read-only, using keyfiles:\n"
 					"truecrypt -m ro -k keyfile1,keyfile2 volume.tc\n"
 					"\n"
 					"Mount a volume without mounting its filesystem:\n"
@@ -1030,11 +1246,16 @@ namespace TrueCrypt
 #ifndef TC_NO_GUI
 				if (Application::GetUserInterfaceType() == UserInterfaceType::Graphic)
 				{
-					wxDialog dialog (nullptr, wxID_ANY, _("TrueCrypt Command Line Help"), wxDefaultPosition, wxSize (600,400));
+					wxDialog dialog (nullptr, wxID_ANY, _("TrueCrypt Command Line Help"), wxDefaultPosition);
 
 					wxTextCtrl *textCtrl = new wxTextCtrl (&dialog, wxID_ANY, wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE | wxTE_READONLY);
+					textCtrl->SetFont (wxFont (wxNORMAL_FONT->GetPointSize(), wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL, false, L"Courier"));
 					textCtrl->SetValue (helpText);
 
+					int fontWidth, fontHeight;
+					textCtrl->GetTextExtent (L"A", &fontWidth, &fontHeight);
+					dialog.SetSize (wxSize (fontWidth * 85, fontHeight * 29));
+
 					wxBoxSizer *sizer = new wxBoxSizer (wxVERTICAL);
 					sizer->Add (textCtrl, 1, wxALL | wxEXPAND, 5);
 					sizer->Add (new wxButton (&dialog, wxID_OK, _("OK")), 0, wxALL | wxALIGN_CENTER_HORIZONTAL, 5);
@@ -1052,6 +1273,18 @@ namespace TrueCrypt
 			}
 			return true;
 
+		case CommandId::ExportSecurityTokenKeyfile:
+			ExportSecurityTokenKeyfile();
+			return true;
+
+		case CommandId::ImportSecurityTokenKeyfiles:
+			ImportSecurityTokenKeyfiles();
+			return true;
+
+		case CommandId::ListSecurityTokenKeyfiles:
+			ListSecurityTokenKeyfiles();
+			return true;
+
 		case CommandId::ListVolumes:
 			if (Preferences.Verbose)
 				DisplayVolumeProperties (cmdLine.ArgVolumes);
@@ -1059,6 +1292,14 @@ namespace TrueCrypt
 				ListMountedVolumes (cmdLine.ArgVolumes);
 			return true;
 
+		case CommandId::RestoreHeaders:
+			RestoreVolumeHeaders (cmdLine.ArgVolumePath);
+			return true;
+
+		case CommandId::SavePreferences:
+			Preferences.Save();
+			return true;
+
 		case CommandId::Test:
 			Test();
 			return true;
@@ -1073,6 +1314,9 @@ namespace TrueCrypt
 	void UserInterface::SetPreferences (const UserPreferences &preferences)
 	{
 		Preferences = preferences;
+
+		Cipher::EnableHwSupport (!preferences.DefaultMountOptions.NoHardwareCrypto);
+
 		PreferencesUpdatedEvent.Raise();
 	}
 
@@ -1123,11 +1367,11 @@ namespace TrueCrypt
 			return wxString::Format (L"%.1f %s", (double)(speed/1024.0/1024/1024/1024), LangString["TB_PER_SEC"].c_str());
 		else if (speed > 1024ULL*1024*1024*99)
 			s << speed/1024/1024/1024 << L" " << LangString["GB_PER_SEC"].c_str();
-		else if (speed > 1024ULL*1024*1024)
+		else if (speed > 1024ULL*1024*999)
 			return wxString::Format (L"%.1f %s", (double)(speed/1024.0/1024/1024), LangString["GB_PER_SEC"].c_str());
-		else if (speed > 1024ULL*1024*99)
+		else if (speed > 1024ULL*1024*9)
 			s << speed/1024/1024 << L" " << LangString["MB_PER_SEC"].c_str();
-		else if (speed > 1024ULL*1024)
+		else if (speed > 1024ULL*999)
 			return wxString::Format (L"%.1f %s", (double)(speed/1024.0/1024), LangString["MB_PER_SEC"].c_str());
 		else if (speed > 1024ULL)
 			s << speed/1024 << L" " << LangString["KB_PER_SEC"].c_str();
